from .GenericSource import *
import sys
import time
import scipy
import threading


class GANSourceDefaultGenerator:
    """
    This class manage the base components of a particle generator.

    - In the constructor, the module 'gaga' is imported. It is only imported in the constructor to only required
    this module if it is used

    - 'initialize' function: the GAN is loaded and the list of keys is initialized

    - 'generator' function: default generator

    - 'get_output_keys' function: map the user defined keys to the ones of the generator. There are two usages, either
    with on single primary (3 values for position, direction), or paired primary (6 values).

    - 'move_backward' function: consider all particles positions and move their backward according to their direction,
    using the factor provided by the user in 'user_info.backward_distance'. This is useful to allow generating
    particles that do not intersect with the detector.

    - 'copy_generated_particle_to_g4' function: copy all the particles (pos, dir, time, energy) to the cpp part.

    """

    def __init__(self, user_info):
        self.user_info = user_info
        self.gaga = gate.import_gaga_phsp()
        self.indexes_are_build = None
        if self.gaga is None:
            print("Cannot run GANSource, gaga_phsp not installed?")
            sys.exit()
        self.lock = threading.Lock()
        self.initialize_is_done = False
        self.keys_output = None
        self.gan_info = None

    def __getstate__(self):
        self.lock = None
        self.gaga = None
        self.gan_info = None
        return self.__dict__

    def initialize(self):
        with self.lock:
            if not self.initialize_is_done:
                self.read_gan_and_keys()
                self.initialize_is_done = True

    def read_gan_and_keys(self):
        # allow converting str like 1e5 to int
        self.user_info.batch_size = int(float(self.user_info.batch_size))

        # FIXME check the number of params

        # read pth and create the gan info structure
        self.gan_info = Box()
        g = self.gan_info
        g.params, g.G, g.D, g.optim, g.dtypef = self.gaga.load(
            self.user_info.pth_filename, "auto", verbose=False
        )

        """
        gan_info structure
        - params    net params from gaga_phsp (including the keys)
        - G         Generator net
        - D         Discriminator net
        - optim     Info about net optimisation
        - dtypef    CPU or GPU

        We analyse the keys and fill the following elements to initialize the GANSource
        - info.position_is_set_by_GAN
        - info.direction_is_set_by_GAN
        - info.energy_is_set_by_GAN
        - info.time_is_set_by_GAN
        - info.weight_is_set_by_GAN
        - info.timing_is_relative
        """

        # by default, the output keys are the same as the input keys
        # (could be changed later by some parameterization)
        if self.keys_output is None:
            g.params.keys_output = g.params.keys_list
        else:
            g.params.keys_output = self.keys_output

        # position,
        g.position_is_set_by_GAN = self.user_info.position_keys is not None
        g.direction_is_set_by_GAN = self.user_info.direction_keys is not None
        g.energy_is_set_by_GAN = self.user_info.energy_key is not None
        g.time_is_set_by_GAN = self.user_info.time_key is not None
        g.weight_is_set_by_GAN = self.user_info.weight_key is not None
        g.timing_is_relative = self.user_info.relative_timing

        # init the key index
        g.position_gan_index = -1
        g.direction_gan_index = -1
        g.energy_gan_index = -1
        g.time_gan_index = -1
        g.weight_gan_index = -1

    def get_output_keys(self):
        with self.lock:
            # (this is only performed once)
            if not self.indexes_are_build:
                self.get_output_keys_with_lock()
                self.indexes_are_build = True

    def get_output_keys_with_lock(self):
        n = self.user_info.batch_size
        g = self.gan_info
        the_keys = g.params.keys_output

        self.check_parameters(g)
        self.get_position_index(g, the_keys, n)
        self.get_direction_index(g, the_keys, n)
        self.get_energy_index(g, the_keys, n)
        self.get_time_index(g, the_keys, n)
        self.get_weight_index(g, the_keys, n)

    def check_parameters(self, g):
        # position
        if g.position_is_set_by_GAN and len(self.user_info.position_keys) != 3:
            dim = len(self.user_info.position_keys)
            self.fatal(f"you must provide 3 values for position, while it was {dim}")
        # direction
        if g.direction_is_set_by_GAN and len(self.user_info.direction_keys) != 3:
            dim = len(self.user_info.direction_keys)
            self.fatal(f"you must provide 3 values for direction, while it was {dim}")

    def fatal(self, txt):
        gate.fatal(f"Error in the GANSource {self.user_info.name}: {txt}")

    def get_position_index(self, g, the_keys, n):
        # get position index from GAN (or a fixed value)
        if not g.position_is_set_by_GAN:
            return
        g.position_gan_index, g.position_use_index = self.get_gan_key_index(
            the_keys, self.user_info.position_keys, n
        )

    def get_direction_index(self, g, the_keys, n):
        # get position from GAN (or a fixed value)
        if not g.direction_is_set_by_GAN:
            return
        g.direction_gan_index, g.direction_use_index = self.get_gan_key_index(
            the_keys, self.user_info.direction_keys, n
        )

    def get_energy_index(self, g, the_keys, n):
        # get energy index from GAN
        if not g.energy_is_set_by_GAN:
            return
        g.energy_gan_index = the_keys.index(self.user_info.energy_key)

    def get_time_index(self, g, the_keys, n):
        # get time index from GAN
        if not g.time_is_set_by_GAN:
            return
        g.time_gan_index = the_keys.index(self.user_info.time_key)

    def get_weight_index(self, g, the_keys, n):
        # get weight index from GAN
        if not g.weight_is_set_by_GAN:
            return
        g.weight_gan_index = the_keys.index(self.user_info.weight_key)

    def get_gan_key_index(self, all_keys, user_keys, n):
        """
        Consider 'user_keys' in the list of all keys and return the indexes
        Special case: if the key name is a float value, we consider the value instead of the index.
        """
        p = []
        o = []  # true or false if used or not
        dim = len(user_keys)
        for i in range(dim):
            try:
                index = all_keys.index(user_keys[i])
                p.append(index)
                o.append(True)
            except:
                if type(user_keys[i]) == float:
                    p.append(np.ones(n) * user_keys[i])
                    o.append(False)
                else:
                    gate.fatal(
                        f"Error, cannot use the key {user_keys[i]} (in {user_keys}) in the GAN source. "
                        f"GAN keys are: {all_keys}"
                    )
        return p, o

    def generator(self, source):
        """
        Main function that will be called from the cpp side every time a batch
        of particles should be created.
        Once created here, the particles are copied to cpp.
        (Yes maybe the copy could be avoided, but I did not manage to do it)
        """
        # get the info
        g = self.gan_info
        n = self.user_info.batch_size
        start = None

        # verbose and timing ?
        if self.user_info.verbose_generator:
            start = time.time()
            print(f"Generate {n} particles from GAN ", end="")

        # generate samples (this is the most time-consuming part)
        fake = self.gaga.generate_samples2(
            g.params,
            g.G,
            g.D,
            n=n,
            batch_size=n,
            normalize=False,
            to_numpy=True,
            silence=True,
        )

        # consider the names of the output keys position/direction/energy/time/weight
        self.get_output_keys()

        # move particle backward ?
        self.move_backward(g, fake)

        # copy to cpp
        self.copy_generated_particle_to_g4(source, g, fake)

        # verbose
        if self.user_info.verbose_generator:
            end = time.time()
            print(f"in {end - start:0.1f} sec (GPU={g.params.current_gpu})")

    def copy_generated_particle_to_g4(self, source, g, fake):
        # get the index of from the GAN vector
        # (or some fixed values)

        # position
        if g.position_is_set_by_GAN:
            pos = []
            dim = len(g.position_gan_index)
            for i in range(dim):
                if g.position_use_index[i]:
                    pos.append(fake[:, g.position_gan_index[i]])
                else:
                    pos.append(g.position_gan_index[i])
            # copy to c++
            source.fPositionX = pos[0]
            source.fPositionY = pos[1]
            source.fPositionZ = pos[2]

        # direction
        if g.direction_is_set_by_GAN:
            dir = []
            dim = len(g.direction_gan_index)
            for i in range(dim):
                if g.direction_use_index[i]:
                    dir.append(fake[:, g.direction_gan_index[i]])
                else:
                    dir.append(g.direction_gan_index[i])
            # copy to c++
            source.fDirectionX = dir[0]
            source.fDirectionY = dir[1]
            source.fDirectionZ = dir[2]

        # energy
        if g.energy_is_set_by_GAN:
            # copy to c++
            source.fEnergy = fake[:, g.energy_gan_index]

        # time
        if g.time_is_set_by_GAN:
            # copy to c++
            source.fTime = fake[:, g.time_gan_index]

        # weight
        if g.weight_is_set_by_GAN:
            # copy to c++
            source.fWeight = fake[:, g.weight_gan_index]

    def move_backward(self, g, fake):
        # move particle backward ?
        back = self.user_info.backward_distance
        if not back:
            return

        if not g.time_is_set_by_GAN and not self.user_info.backward_force:
            gate.fatal(
                f"If backward is enabled the time is not managed by GAN,"
                f" time is wrong. IT can be forced, however, with the option 'backward_force'"
            )

        # move particle position
        position = fake[:, g.position_gan_index[0] : g.position_gan_index[0] + 3]
        direction = fake[:, g.direction_gan_index[0] : g.direction_gan_index[0] + 3]
        fake[:, g.position_gan_index[0] : g.position_gan_index[0] + 3] = (
            position - back * direction
        )

        # modify the time because we move the particle backward
        if g.time_is_set_by_GAN:
            c = scipy.constants.speed_of_light * 1000 / 1e9  # in mm/ns
            xt = back / c
            fake[:, g.time_gan_index] -= xt



import opengate_core as g4
from ..UserElement import *
from scipy.spatial.transform import Rotation
from box import BoxList
import time


class VolumeBase(UserElement):
    """
    Store information about a geometry volume:
    - G4 objects: Solid, LogicalVolume, PhysicalVolume
    - user parameters: user_info
    - additional data such as: mother, material etc
    """

    @staticmethod
    def set_default_user_info(user_info):
        gate.UserElement.set_default_user_info(user_info)
        user_info.mother = gate.__world_name__
        user_info.material = "G4_AIR"
        user_info.translation = [0, 0, 0]
        user_info.color = [1, 1, 1, 1]
        user_info.rotation = Rotation.identity().as_matrix()
        user_info.repeat = None
        user_info.build_physical_volume = True

    def __init__(self, user_info):
        super().__init__(user_info)
        # convert the list of repeat to a BoxList to easier access
        self.user_info.repeat = BoxList(self.user_info.repeat)
        # init
        self.g4_solid = None
        self.g4_logical_volume = None
        self.g4_vis_attributes = None
        # one volume may have several physical volume, this is the first one:
        self.g4_physical_volume = None
        # this list is all volumes (including first)
        self.g4_physical_volumes = []
        self.material = None
        self.g4_region = None
        # used
        self.volume_engine = None

    def __del__(self):
        pass

    def __str__(self):
        s = f"Volume: {self.user_info}"
        return s

    def build_solid(self):
        gate.fatal(f'Need to overwrite "build_solid" in {self.user_info}')

    def construct(self, volume_engine):
        self.volume_engine = volume_engine
        # check placements
        ui = self.user_info
        if ui.repeat:
            if ui.translation is not None or ui.rotation is not None:
                gate.fatal(
                    f'When using "repeat", translation and rotation must be None, '
                    f"for volume : {ui.name}"
                )
        # construct solid/material/lv/pv/regions
        self.construct_solid()
        self.construct_material(volume_engine)
        self.construct_logical_volume()
        if self.user_info.build_physical_volume:
            self.construct_physical_volume()
        self.construct_region()

    def construct_solid(self):
        # builder the G4 solid
        self.g4_solid = self.build_solid()
        # build the solid according to the type
        # self.g4_solid = self.solid_builder.Build(self.user_info)

    def construct_material(self, volume_engine):
        # retrieve or build the material
        self.material = volume_engine.find_or_build_material(self.user_info.material)

    def construct_logical_volume(self):
        self.g4_logical_volume = g4.G4LogicalVolume(
            self.g4_solid, self.material, self.user_info.name  # solid  # material
        )  # name
        # color
        self.g4_vis_attributes = g4.G4VisAttributes()
        self.g4_vis_attributes.SetColor(*self.user_info.color)
        if self.user_info.color[3] == 0:
            self.g4_vis_attributes.SetVisibility(False)
        else:
            self.g4_vis_attributes.SetVisibility(True)
        self.g4_logical_volume.SetVisAttributes(self.g4_vis_attributes)

    def construct_physical_volume(self):
        # find the mother's logical volume
        if self.user_info.mother:
            st = g4.G4LogicalVolumeStore.GetInstance()
            mother_logical = st.GetVolume(self.user_info.mother, False)
        else:
            mother_logical = None
        # consider the 3D transform -> helpers_transform.
        if self.user_info.repeat:
            self.construct_physical_volume_repeat(mother_logical)
        else:
            transform = gate.get_vol_g4_transform(self.user_info)
            check = self.volume_engine.simulation.user_info.check_volumes_overlap
            self.g4_physical_volume = g4.G4PVPlacement(
                transform,
                self.g4_logical_volume,  # logical volume
                self.user_info.name,  # volume name
                mother_logical,  # mother volume or None if World
                False,  # no boolean operation # FIXME for BooleanVolume ?
                0,  # copy number
                check,
            )  # overlaps checking
            self.g4_physical_volumes.append(self.g4_physical_volume)

    def construct_physical_volume_repeat(self, mother_logical):
        check = self.volume_engine.simulation.user_info.check_volumes_overlap
        i = 0
        for repeat_vol in self.user_info.repeat:
            transform = gate.get_vol_g4_transform(repeat_vol)
            v = g4.G4PVPlacement(
                transform,
                self.g4_logical_volume,  # logical volume
                repeat_vol.name,  # volume name
                mother_logical,  # mother volume or None if World
                False,  # no boolean operation
                i,  # copy number
                check,
            )  # overlaps checking
            i += 1
            self.g4_physical_volumes.append(v)
        self.g4_physical_volume = self.g4_physical_volumes[0]

    def construct_region(self):
        if self.user_info.name == gate.__world_name__:
            # the default region for the world is set by G4 RunManagerKernel
            return
        rs = g4.G4RegionStore.GetInstance()
        self.g4_region = rs.FindOrCreateRegion(self.user_info.name)
        # set a fake default production cuts to avoid warning
        # (warning in G4RunManagerKernel::CheckRegions())
        # keep it in self to avoid garbage collecting
        self.fake_cuts = g4.G4ProductionCuts()
        self.g4_region.SetProductionCuts(self.fake_cuts)
        # set region and Log Vol
        self.g4_logical_volume.SetRegion(self.g4_region)
        self.g4_region.AddRootLogicalVolume(self.g4_logical_volume, True)



import colored
import numpy as np
import opengate as gate
import opengate_core as g4
from box import Box
import textwrap
from inspect import getframeinfo, stack
import pkg_resources
import sys
from pathlib import Path
import random
import string
import os
from numpy.random import MT19937
from numpy.random import RandomState, SeedSequence
import inspect
import re
import json
from importlib.metadata import version
import git

color_error = colored.fg("red") + colored.attr("bold")
color_warning = colored.fg("orange_1")
color_ok = colored.fg("green")


def fatal(s):
    caller = getframeinfo(stack()[1][0])
    ss = f"(in {caller.filename} line {caller.lineno})"
    ss = colored.stylize(ss, color_error)
    gate.log.critical(ss)
    s = colored.stylize(s, color_error)
    gate.log.critical(s)
    sys.exit(-1)


def warning(s):
    s = colored.stylize(s, color_warning)
    gate.log.warning(s)


def raise_except(s):
    s = colored.stylize(s, color_error)
    raise Exception(s)


def assert_equal_dic(d1, d2, name=""):
    for k in d1:
        if not k in d2:
            fatal(f"ERROR missing key {k} in {name}")
        if isinstance(d1[k], np.ndarray):
            if np.any(d2[k] != d1[k]):
                fatal(f"ERROR np array {k} {d1[k]} in {name}")
        else:
            if d2[k] != d1[k]:
                fatal(f"ERROR value for {k} in {name}")
    for k in d2:
        if not k in d1:
            fatal(f"ERROR, additional key {k} in {name}")


def g4_units(name: str) -> float:
    table = g4.G4UnitDefinition.GetUnitsTable()
    for t in table:
        for a in t.GetUnitsList():
            if a.GetName() == name or a.GetSymbol() == name:
                return a.GetValue()
    units_list = []
    for t in table:
        for a in t.GetUnitsList():
            units_list.append(a.GetSymbol())
    s = [str(u) + " " for u in units_list]
    fatal(f"Error, cannot find the unit named {name}. Known are: {s}")


def g4_best_unit(value, unit_type):
    return g4.G4BestUnit(value, unit_type)


def assert_key(key: str, d: Box):
    if key not in d:
        gate.fatal(f'The key "{key}" is needed in this structure:\n' f"{d}")


def assert_keys(keys: list, d: Box):
    for key in keys:
        assert_key(key, d)


def indent(amount, text, ch=" "):
    """
    Prefix the text with indent spaces
    https://stackoverflow.com/questions/8234274/how-to-indent-the-contents-of-a-multi-line-string
    """
    return textwrap.indent(text, amount * ch)


def assert_unique_element_name(elements, name):
    if name in elements:
        s = (
            f"Error, cannot add '{name}' because this element's name already exists"
            f" in: {elements}."
        )
        gate.fatal(s)


def make_builders(class_names):
    """
    Consider a list of Classname. For each, it build a key/value, with:
    - the type of the class as key
    - and a lambda function that create an object of this class as value
    """
    builder_list = {}
    for c in class_names:
        # note the following lambda:
        # https://stackoverflow.com/questions/2295290/what-do-lambda-function-closures-capture
        try:
            builder_list[c.type_name] = lambda x, y=c: y(x)
        except AttributeError:
            # if type_name is not an attribute of the class,
            # we use the name of the class as key.
            # Also: no name parameter (this is for Physics List)
            builder_list[c.__name__] = lambda y=c: y()
    return builder_list


def read_mac_file_to_commands(filename):
    # read a file located into the 'mac' folder of the source code
    # return a list of commands
    resource_package = __name__
    resource_path = "/".join(("mac", filename))  # Do not use os.path.join()
    template = pkg_resources.resource_string(resource_package, resource_path)
    c = template.decode("utf-8")
    commands = []
    for s in c.split("\n"):
        if s == "":
            continue
        # if s[0] == '#':
        #    continue
        commands.append(s)
    return commands


def check_filename_type(filename):
    # Algorithms (itk) do not support Path -> convert to str
    if isinstance(filename, Path):
        return str(filename)
    return filename


def get_random_folder_name(size=8, create=True):
    r = "".join(random.choices(string.ascii_lowercase + string.digits, k=size))
    r = "run." + r
    if create:
        if not os.path.exists(r):
            print(f"Creating output folder {r}")
            os.mkdir(r)
        if not os.path.isdir(r):
            gate.fatal(f"Error, while creating {r}.")
    return r


def import_gaga_phsp():
    # Try to import torch
    try:
        import torch
    except:
        gate.fatal(
            f'The module "torch" is needed, see https://pytorch.org/get-started/locally/ to install it'
        )

    # Try to import gaga_phsp
    try:
        import gaga_phsp as gaga
    except:
        gate.fatal("The module \"gaga_phsp\" is needed. Use 'pip install gaga_phsp'")

    # Check minimal version of gaga_phsp
    import pkg_resources
    from packaging import version

    gaga_version = pkg_resources.get_distribution("gaga_phsp").version
    gaga_minimal_version = "0.5.8"
    if version.parse(gaga_version) < version.parse(gaga_minimal_version):
        gate.fatal(
            "The minimal version of gaga_phsp is not correct. You should install at least the version "
            + gaga_minimal_version
            + ". Your version is "
            + gaga_version
        )
    return gaga


def import_garf():
    # Try to import torch
    try:
        import torch
    except:
        gate.fatal(
            f'The module "torch" is needed, see https://pytorch.org/get-started/locally/ to install it'
        )

    # Try to import garf_phsp
    try:
        import garf
    except:
        gate.fatal("The module \"garf\" is needed. Use ' pip install garf'")

    # Check minimal version of garf
    import pkg_resources
    from packaging import version

    garf_version = pkg_resources.get_distribution("garf").version
    garf_minimal_version = "2.2"
    if version.parse(garf_version) < version.parse(garf_minimal_version):
        gate.fatal(
            "The minimal version of garf is not correct. You should install at least the version "
            + garf_minimal_version
            + ". Your version is "
            + garf_version
        )
    return garf


def get_rnd_seed(seed):
    return RandomState(MT19937(SeedSequence(seed)))


def DDF():
    """
    Debug print current Function name
    """
    print("--> Entering", inspect.stack()[1][3])


def DD(arg):
    """
    Debug print variable name and its value
    """
    frame = inspect.currentframe()
    try:
        context = inspect.getframeinfo(frame.f_back).code_context
        caller_lines = "".join([line.strip() for line in context])
        m = re.search(r"DD\s*\((.+?)\);*$", caller_lines)
        if m:
            caller_lines = m.group(1)
            # end if
        print(caller_lines, "=", arg)
    finally:
        del frame


def print_dic(dic):
    print(json.dumps(dic, indent=4, default=str))


def get_release_date(opengate_version):
    import requests

    package_name = "opengate"
    url = f"https://pypi.org/pypi/{package_name}/json"
    response = requests.get(url)

    if response.status_code == 200:
        package_data = response.json()
        releases = package_data["releases"]
        if releases:
            # latest_version = max(releases, key=lambda v: package_data['releases'][v][0]['upload_time'])
            release_date = package_data["releases"][opengate_version][0]["upload_time"]
            return f"{release_date}"
        else:
            return "unknown"
    else:
        return "unknown"


def print_opengate_info():
    """
    Print information about OpenGate and the environment
    """

    gi = g4.GateInfo
    v = gi.get_G4Version().replace("$Name: ", "")
    v = v.replace("$", "")

    pv = sys.version.replace("\n", "")
    print(f"Python version   {pv}")
    print(f"Platform         {sys.platform}")
    print(f"Site package     {g4.get_site_packages_dir()}")

    print(f"Geant4 version   {v}")
    print(f"Geant4 MT        {gi.get_G4MULTITHREADED()}")
    print(f"Geant4 GDML      {gi.get_G4GDML()}")
    print(f"Geant4 date      {gi.get_G4Date()}")
    print(f"Geant4 data      {g4.get_G4_data_folder()}")

    print(f"ITK version      {gi.get_ITKVersion()}")

    print(f"GATE version     {version('opengate')}")
    print(f"GATE folder      {gate.__path__[0]}")

    # check if from a git version ?
    git_path = Path(gate.__path__[0]) / ".."
    try:
        git_repo = git.Repo(git_path)
        sha = git_repo.head.object.hexsha
        print(f"GATE git sha     {sha}")
        commit = git_repo.head.commit
        commit_date = commit.committed_datetime
        print(f"GATE date        {commit_date}  (last commit)")

    except:
        print(f"GATE date        {get_release_date(version('opengate'))} (pypi)")



import opengate_core as g4
import opengate as gate
import sys


class KineticEnergyFilter(g4.GateKineticEnergyFilter, gate.UserElement):
    type_name = "KineticEnergyFilter"

    def set_default_user_info(user_info):
        gate.UserElement.set_default_user_info(user_info)
        # required user info, default values
        user_info.energy_min = 0
        user_info.energy_max = sys.float_info.max

    def __init__(self, user_info):
        g4.GateKineticEnergyFilter.__init__(self)  # no argument in cpp side
        gate.UserElement.__init__(self, user_info)
        # type_name MUST be defined in class that inherit from a Filter

    def __getstate__(self):
        # needed to not pickle the g4.GateKineticEnergyFilter
        return self.__dict__



import opengate as gate
import opengate_core as g4
import numpy as np
import itk
from scipy.spatial.transform import Rotation


class DigitizerProjectionActor(g4.GateDigitizerProjectionActor, gate.ActorBase):
    """
    This actor takes as input HitsCollections and performed binning in 2D images.
    If there are several HitsCollection as input, the slices will correspond to each HC.
    If there are several runs, images will also be slice-stacked.
    """

    type_name = "HitsProjectionActor"

    @staticmethod
    def set_default_user_info(user_info):
        gate.ActorBase.set_default_user_info(user_info)
        mm = gate.g4_units("mm")
        user_info.output = False
        user_info.input_digi_collections = ["Hits"]
        user_info.spacing = [4 * mm, 4 * mm]
        user_info.size = [128, 128]
        user_info.physical_volume_index = None
        user_info.origin_as_image_center = True
        user_info.detector_orientation_matrix = Rotation.from_euler("x", 0).as_matrix()

    def __init__(self, user_info):
        gate.ActorBase.__init__(self, user_info)
        g4.GateDigitizerProjectionActor.__init__(self, user_info.__dict__)
        actions = {"StartSimulationAction", "EndSimulationAction"}
        self.AddActions(actions)
        self.output_image = None
        if len(user_info.input_digi_collections) < 1:
            gate.fatal(f"Error, not input hits collection.")

    def __del__(self):
        pass

    def __str__(self):
        s = f"DigitizerProjectionActor {self.user_info.name}"
        return s

    def __getstate__(self):
        gate.ActorBase.__getstate__(self)
        self.output_image = None
        return self.__dict__

    def compute_thickness(self, volume, channels):
        """
        Get the thickness of the detector volume, in the correct direction.
        By default, it is Z. We use the 'projection_orientation' to get the correct one.
        """
        vol = self.volume_engine.get_volume(volume)
        solid = vol.g4_physical_volumes[0].GetLogicalVolume().GetSolid()
        pMin = g4.G4ThreeVector()
        pMax = g4.G4ThreeVector()
        solid.BoundingLimits(pMin, pMax)
        d = np.array([0, 0, 1.0])
        d = np.dot(self.user_info.detector_orientation_matrix, d)
        imax = np.argmax(d)
        thickness = (pMax[imax] - pMin[imax]) / channels
        return thickness

    def StartSimulationAction(self):
        # check size and spacing
        if len(self.user_info.size) != 2:
            gate.fatal(f"Error, the size must be 2D while it is {self.user_info.size}")
        if len(self.user_info.spacing) != 2:
            gate.fatal(
                f"Error, the spacing must be 2D while it is {self.user_info.spacing}"
            )
        self.user_info.size.append(1)
        self.user_info.spacing.append(1)

        # for the moment, we cannot use this actor with several volumes
        m = self.user_info.mother
        if hasattr(m, "__len__") and not isinstance(m, str):
            gate.fatal(
                f"Sorry, cannot (yet) use several mothers volumes for "
                f"DigitizerProjectionActor {self.user_info.name}"
            )

        # define the new size and spacing according to the nb of channels
        # and according to the volume shape
        size = np.array(self.user_info.size)
        spacing = np.array(self.user_info.spacing)
        size[2] = len(self.user_info.input_digi_collections) * len(
            self.simulation.run_timing_intervals
        )
        spacing[2] = self.compute_thickness(self.user_info.mother, size[2])

        # create image
        self.output_image = gate.create_3d_image(size, spacing)

        # initial position (will be anyway updated in BeginOfRunSimulation)
        pv = None
        try:
            pv = gate.get_physical_volume(
                self.volume_engine,
                self.user_info.mother,
                self.user_info.physical_volume_index,
            )
        except:
            gate.fatal(f"Error in the HitsProjectionActor {self.user_info.name}")
        gate.attach_image_to_physical_volume(pv.GetName(), self.output_image)
        self.fPhysicalVolumeName = str(pv.GetName())
        # update the cpp image and start
        gate.update_image_py_to_cpp(self.output_image, self.fImage, True)
        g4.GateDigitizerProjectionActor.StartSimulationAction(self)

    def EndSimulationAction(self):
        g4.GateDigitizerProjectionActor.EndSimulationAction(self)
        # retrieve the image
        self.output_image = gate.get_cpp_image(self.fImage)
        info = gate.get_info_from_image(self.output_image)
        # change the
        # g and origin for the third dimension
        spacing = self.output_image.GetSpacing()
        origin = self.output_image.GetOrigin()
        # should we center the projection ?
        if self.user_info.origin_as_image_center:
            origin = -info.size * info.spacing / 2.0 + info.spacing / 2.0
        spacing[2] = 1
        origin[2] = 0
        self.output_image.SetSpacing(spacing)
        self.output_image.SetOrigin(origin)
        if self.user_info.output:
            itk.imwrite(
                self.output_image, gate.check_filename_type(self.user_info.output)
            )



import opengate as gate
import opengate_core as g4
import re
from box import Box


class MaterialBuilder:
    """
    A description of a material, that will can be build on demand.
    A material is described by a list of components that can be elements or sub-materials.
    """

    def __init__(self, material_database):
        self.type = "material"
        self.name = None
        self.symbol = None
        self.density = None
        self.n = None
        self.state = None
        self.components = {}
        self.material_database = material_database

    def __del__(self):
        pass

    def __repr__(self):
        s = f"({self.type}) {self.name} {self.density} {self.n} {self.components}"
        return s

    def read(self, f, line):
        # read the name
        s = line.split(":")
        if len(s) != 2:
            gate.fatal(
                f"Error line {line}, expecting a material name follow by a colon ':'."
            )
        name = s[0]
        self.name = name

        # reading density, n, state
        s = s[1].split(";")
        if len(s) != 3 and len(s) != 2:
            gate.fatal(f"Error while parsing material {self.name}, line {line}")

        # density
        self.density = gate.read_tag_with_unit(s[0], "d")
        if not self.density:
            gate.fatal(
                f"Error while parsing material {self.name}, line {line}\n"
                f'Expected density with "d=XXX"'
            )

        # nb of components
        self.n = int(gate.read_tag(s[1], "n"))

        # state
        if len(s) > 2:
            self.state = gate.read_tag(s[2], "state")
            if self.state:
                self.state = self.state.lower()

        # elements
        for e in range(self.n):
            line = gate.read_next_line(f)
            if line.startswith("+mat:"):
                e = self.read_one_submat(line)
                self.components[e.name] = e
            if line.startswith("+el"):
                e = self.read_one_element(line)
                self.components[e.name] = e

    def read_one_element(self, line):
        # skip the initial +el
        s = line.split("+el:")
        s = re.split("[;,]", s[1])
        if len(s) != 2:
            gate.fatal(
                f"Error while reading the line: {line} \n"
                f'Expected "name=" ; "n=" or "f="'
            )
        # read the name
        elname = gate.read_tag(s[0], "name")
        if elname == "auto":
            elname = self.name
        if not elname:
            gate.fatal(
                f"Error reading line {line} \n during the elements of material {self.name}"
            )
        # read f or n, put the other one to 'None'
        f = None
        n = gate.read_tag(s[1], "n")
        if not n:
            f = float(gate.read_tag(s[1], "f"))
        else:
            n = int(n)
        e = Box({"name": elname, "n": n, "f": f, "type": "element"})
        return e

    def read_one_submat(self, line):
        # skip the initial +mat
        s = line.split("+mat:")
        s = re.split("[;,]", s[1])
        if len(s) != 2:
            gate.fatal(
                f"Error while reading the line: {line} \n"
                f'Expected "name=" ; "n=" or "f="'
            )
        # read the name
        elname = gate.read_tag(s[0], "name")
        if not elname:
            gate.fatal(
                f"Error reading line {line} \n during the elements of material {self.name}"
            )
        # read f
        f = gate.read_tag(s[1], "f")
        if f is not None:
            f = float(f)
        else:
            gate.fatal(
                f"Error during reading material database {self.material_database.current_filename}"
                f", for the sub material {elname}, except fraction 'f=', while the line is {line}"
            )

        # build the dict
        e = Box({"name": elname, "n": None, "f": f, "type": "material"})
        return e

    def build(self):
        switcher = {
            None: g4.G4State.kStateUndefined,
            "solid": g4.G4State.kStateSolid,
            "liquid": g4.G4State.kStateLiquid,
            "gaz": g4.G4State.kStateGas,
            "gas": g4.G4State.kStateGas,
        }
        state = switcher.get(self.state, f"Invalid material state {self.state}")

        # default temp
        kelvin = gate.g4_units("kelvin")
        temp = 293.15 * kelvin

        # default pressure
        atmosphere = gate.g4_units("atmosphere")
        pressure = 1 * atmosphere

        # compute the correct nb of elements
        n = 0
        for elem in self.components.values():
            if elem.type == "element":
                n += 1
            else:
                subm = self.material_database.FindOrBuildMaterial(elem.name)
                n += len(subm.GetElementVector())

        # create material
        mat = g4.G4Material(self.name, self.density, n, state, temp, pressure)

        # add all components
        for elem in self.components.values():
            if elem.type == "element":
                self.add_element_to_material(mat, elem)
            else:
                self.add_submat_to_material(mat, elem)
        return mat

    def add_element_to_material(self, mat, elem):
        b = self.material_database.FindOrBuildElement(elem.name)
        if elem.f is None:
            mat.AddElementByNumberOfAtoms(b, elem.n)  # FIXME AddElementByMassFraction
        else:
            mat.AddElementByMassFraction(b, elem.f)

    def add_submat_to_material(self, mat, elem):
        subm = self.material_database.FindOrBuildMaterial(elem.name)
        elems = subm.GetElementVector()
        i = 0
        for el in elems:
            elf = subm.GetElementFraction(i)
            f = elf * elem.f
            name = str(el.GetName())
            b = self.material_database.FindOrBuildElement(name)
            mat.AddElementByMassFraction(b, f)
            i += 1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test049_pet_digit_blurring_helpers import *

paths = gate.get_default_test_paths(__file__, "gate_test049_pet_blur")

"""
see https://github.com/teaghan/PET_MonteCarlo
and https://doi.org/10.1002/mp.16032

PET simulation to test blurring options of the digitizer

- PET:
- phantom: nema necr
- output: singles with and without various blur options
"""

# create the simulation
sim = gate.Simulation()
nb_threads = 2
create_simulation(sim, nb_threads)

# start simulation
output = sim.start()

# print results
stats = output.get_actor("Stats")
print(stats)

# ----------------------------------------------------------------------------------------------------------
readout = output.get_actor("Singles")
ig = readout.GetIgnoredHitsCount()
print()
print(f"Nb of ignored hits : {ig}")

# check stats
print()
gate.warning(f"Check stats")
p = paths.gate_output
stats_ref = gate.read_stat_file(p / "stats.txt")
stats_ref.counts.run_count = nb_threads
is_ok = gate.assert_stats(stats, stats_ref, 0.025)

# check root hits
hc = output.get_actor("Hits").user_info
f = p / "pet.root"
is_ok = check_root_hits(paths, 1, f, hc.output, "test049_hits_v2_MT.png") and is_ok

# check root singles
sc = output.get_actor("Singles").user_info
is_ok = (
    check_root_singles(paths, 1, f, sc.output, png_output="test049_singles_v2_MT.png")
    and is_ok
)

# gate.delete_run_manager_if_needed(sim) # no
gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation
import pathlib

pathFile = pathlib.Path(__file__).parent.resolve()

# create the simulation
sim = gate.Simulation()
print(f"Volumes types: {sim.dump_volume_types()}")

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False

# add a material database
sim.add_material_database(pathFile / ".." / "data" / "GateMaterials.db")

#  change world size
m = gate.g4_units("m")
world = sim.world
world.size = [1.5 * m, 1.5 * m, 1.5 * m]

# add a simple volume
waterbox = sim.add_volume("Box", "Waterbox")
cm = gate.g4_units("cm")
waterbox.size = [60 * cm, 60 * cm, 60 * cm]
waterbox.translation = [0 * cm, 0 * cm, 35 * cm]
waterbox.material = "G4_WATER"
waterbox.color = [0, 0, 1, 1]  # blue

# another (child) volume with rotation
mm = gate.g4_units("mm")
sheet = sim.add_volume("Box", "Sheet")
sheet.size = [30 * cm, 30 * cm, 2 * mm]
sheet.mother = "Waterbox"
sheet.material = "Lead"
r = Rotation.from_euler("x", 33, degrees=True)
center = [0 * cm, 0 * cm, 10 * cm]
t = gate.get_translation_from_rotation_with_center(r, center)
sheet.rotation = r.as_matrix()
sheet.translation = t + [0 * cm, 0 * cm, -18 * cm]
sheet.color = [1, 0, 0, 1]  # red

# A sphere
sph = sim.add_volume("Sphere", "mysphere")
sph.rmax = 5 * cm
sph.mother = "Waterbox"
sph.translation = [0 * cm, 0 * cm, -8 * cm]
sph.material = "Lung"
sph.color = [0.5, 1, 0.5, 1]  # kind of green
sph.toto = "nothing"  # ignored, should raise a warning

# A ...thing ?
trap = sim.add_volume("Trap", "mytrap")
trap.mother = "Waterbox"
trap.translation = [0, 0, 15 * cm]
trap.material = "G4_LUCITE"

# default source for tests
source = sim.add_source("GenericSource", "Default")
MeV = gate.g4_units("MeV")
Bq = gate.g4_units("Bq")
source.particle = "proton"
source.energy.mono = 240 * MeV
source.position.radius = 1 * cm
source.direction.type = "momentum"
source.direction.momentum = [0, 0, 1]
source.activity = 2500 * Bq

# add stat actor
sim.add_actor("SimulationStatisticsActor", "Stats")

# run timing
sec = gate.g4_units("second")
sim.run_timing_intervals = [
    [0, 0.5 * sec]
    # ,[0.5 * sec, 1.2 * sec]
]

print(sim)


def check_mat(se):
    # explicit check overlap (already performed during initialize)
    print("check overlap with verbose")
    se.check_volumes_overlap(verbose=True)

    # print info material db
    dbn = sim.dump_material_database_names()
    mnist = se.volume_engine.get_database_material_names("NIST")
    mdb = se.volume_engine.get_database_material_names(
        pathFile / ".." / "data" / "GateMaterials.db"
    )
    dm = se.volume_engine.dump_build_materials()
    print("Material info:")
    print("\t databases    :", dbn)
    print("\t mat in NIST  :", len(mnist), mnist)
    print("\t mat in db    :", mdb)
    print("\t defined mat  :", dm)

    print("dbn", dbn)
    assert dbn == [pathFile / ".." / "data" / "GateMaterials.db"]
    # assert len(mnist) == 308  # Geant4 11.02
    assert len(mnist) == 309  # Geant4 11.1
    assert mdb == [
        "Vacuum",
        "Aluminium",
        "Uranium",
        "Silicon",
        "Germanium",
        "Yttrium",
        "Gadolinium",
        "Lutetium",
        "Tungsten",
        "Lead",
        "Bismuth",
        "NaI",
        "NaITl",
        "PWO",
        "BGO",
        "LSO",
        "Plexiglass",
        "GSO",
        "LuAP",
        "YAP",
        "Water",
        "Quartz",
        "Breast",
        "Air",
        "Glass",
        "Scinti-C9H10",
        "LuYAP-70",
        "LuYAP-80",
        "Plastic",
        "CZT",
        "Lung",
        "Polyethylene",
        "PVC",
        "SS304",
        "PTFE",
        "LYSO",
        "Body",
        "Muscle",
        "LungMoby",
        "SpineBone",
        "RibBone",
        "Adipose",
        "Blood",
        "Heart",
        "Kidney",
        "Liver",
        "Lymph",
        "Pancreas",
        "Intestine",
        "Skull",
        "Cartilage",
        "Brain",
        "Spleen",
        "Testis",
        "PMMA",
    ]
    assert dm == ["G4_AIR", "G4_WATER", "Lead", "Lung", "G4_LUCITE"]
    assert len(se.volume_engine.dump_build_materials()) == 5


# verbose
sim.apply_g4_command("/tracking/verbose 0")
# sim.g4_com("/run/verbose 2")
# sim.g4_com("/event/verbose 2")
# sim.g4_com("/tracking/verbose 1")

# start simulation
se = gate.SimulationEngine(sim)
se.user_fct_after_init = check_mat
output = se.start()

# print results at the end
stats = output.get_actor("Stats")
print(stats)

# check
stats_ref = gate.SimulationStatisticsActor()
c = stats_ref.counts
c.run_count = 1
c.event_count = 1280
c.track_count = 17034  # 25668
c.step_count = 78096  # 99465
# stats_ref.pps = 506.6
sec = gate.g4_units("second")
c.duration = 2.5267 * sec
print("-" * 80)
is_ok = gate.assert_stats(stats, stats_ref, 0.15)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate_core as g4
from scipy.spatial.transform import Rotation
import test036_adder_depth_helpers as t036
import uproot
import matplotlib.pyplot as plt

paths = gate.get_default_test_paths(__file__, "gate_test036_adder_depth")

# create and run the simulation
sim = t036.create_simulation("repeat")

# start simulation
output = sim.start()

# test the output
is_ok = t036.test_output(output)

gate.test_ok(is_ok)



import opengate as gate
import opengate_core as g4
import numpy as np

sigma_to_fwhm = 2 * np.sqrt(2 * np.log(2))
fwhm_to_sigma = 1.0 / sigma_to_fwhm


class DigitizerBlurringActor(g4.GateDigitizerBlurringActor, gate.ActorBase):
    """
    Digitizer module for blurring an attribute (single value only, not a vector).
    Usually for energy or time.
    """

    type_name = "DigitizerBlurringActor"

    @staticmethod
    def set_default_user_info(user_info):
        gate.ActorBase.set_default_user_info(user_info)
        user_info.attributes = []
        user_info.output = "singles.root"
        user_info.input_digi_collection = "Hits"
        user_info.skip_attributes = []
        user_info.clear_every = 1e5
        user_info.blur_attribute = None
        user_info.blur_method = "Gaussian"
        user_info.blur_fwhm = None
        user_info.blur_sigma = None
        user_info.blur_reference_value = None
        user_info.blur_resolution = None
        user_info.blur_slope = None

    def __init__(self, user_info):
        # check and adjust parameters
        self.set_param(user_info)
        # base classes
        gate.ActorBase.__init__(self, user_info)
        g4.GateDigitizerBlurringActor.__init__(self, user_info.__dict__)
        actions = {"StartSimulationAction", "EndSimulationAction"}
        self.AddActions(actions)

    def set_param(self, user_info):
        am = ["Gaussian", "InverseSquare", "Linear"]
        m = user_info.blur_method
        if m not in am:
            gate.fatal(
                f"Error, the blur_method must be within {am}, while it is {user_info.blur_method}"
            )
        if m == "Gaussian":
            self.set_param_gauss(user_info)
        if m == "InverseSquare":
            self.set_param_inverse_square(user_info)
        if m == "Linear":
            self.set_param_linear(user_info)

    def set_param_gauss(self, user_info):
        if user_info.blur_fwhm is not None and user_info.blur_sigma is not None:
            gate.fatal(
                f"Error, use blur_sigma or blur_fwhm, not both "
                f"(there are: {user_info.blur_sigma} and {user_info.blur_fwhm}"
            )
        if user_info.blur_fwhm is not None:
            user_info.blur_sigma = user_info.blur_fwhm * fwhm_to_sigma
        if user_info.blur_sigma is None:
            gate.fatal(f"Error, use blur_sigma or blur_fwhm")
        user_info.blur_reference_value = -1
        user_info.blur_resolution = -1
        user_info.blur_slope = 0

    def set_param_inverse_square(self, user_info):
        if user_info.blur_reference_value < 0 or user_info.blur_reference_value is None:
            gate.fatal(
                f"Error, use positive blur_reference_value "
                f"(current value =  {user_info.blur_reference_value}"
            )
        if user_info.blur_resolution < 0 or user_info.blur_resolution is None:
            gate.fatal(
                f"Error, use positive blur_resolution "
                f"(current value =  {user_info.blur_resolution}"
            )
        user_info.blur_fwhm = -1
        user_info.blur_sigma = -1
        if user_info.blur_slope is None:
            user_info.blur_slope = 0

    def set_param_linear(self, user_info):
        self.set_param_inverse_square(user_info)
        if user_info.blur_slope is None:
            gate.fatal(
                f"Error, use positive blur_slope "
                f"(current value =  {user_info.blur_slope}"
            )

    def __del__(self):
        pass

    def __str__(self):
        s = f"DigitizerBlurringActor {self.user_info.name}"
        return s

    def StartSimulationAction(self):
        g4.GateDigitizerBlurringActor.StartSimulationAction(self)

    def EndSimulationAction(self):
        g4.GateDigitizerBlurringActor.EndSimulationAction(self)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import sys
import pathlib
import os

pathFile = pathlib.Path(__file__).parent.resolve()

cmd = "python " + os.path.join(pathFile, "test022_half_life.py") + " 3"
r = os.system(cmd)

sys.exit(r)



from box import Box
import opengate as gate
import opengate_core as g4
from anytree import Node

""" Global name for the world volume"""
__world_name__ = "world"


class VolumeManager:
    """
    Implementation of G4VUserDetectorConstruction.
    In 'Construct' function, build all volumes in the scene.
    Keep a list of solid, logical volumes, physical volumes, materials.
    """

    def __init__(self, simulation):
        """
        Class that store geometry description.
        """
        self.simulation = simulation

        # list of all user_info describing the volumes
        self.user_info_volumes = {}  # user info only

        # database of materials
        self.material_database = gate.MaterialDatabase()

        # FIXME maybe store solids ?

    def __del__(self):
        # print("del volume manager")
        pass

    def __str__(self):
        s = f"{len(self.user_info_volumes)} volumes"
        return s

    def __getstate__(self):
        """
        This is important : to get actor's outputs from a simulation run in a separate process,
        the class must be serializable (pickle).
        The g4 material databases and the info_volume containing volume from solid have to be removed first.
        """
        self.material_database = {}
        self.user_info_volumes = {}
        return self.__dict__

    def get_volume_user_info(self, name):
        if name not in self.user_info_volumes:
            gate.fatal(
                f"The volume {name} is not in the current "
                f"list of volumes: {self.user_info_volumes}"
            )
        return self.user_info_volumes[name]

    def new_solid(self, solid_type, name):
        if solid_type == "Boolean":
            gate.fatal(f"Cannot create solid {solid_type}")
        # Create a UserInfo for a volume
        u = gate.UserInfo("Volume", solid_type, name)
        # remove unused keys: object, etc. (it's a solid, not a volume)
        VolumeManager._pop_keys_unused_by_solid(u)
        return u

    def get_solid_info(self, user_info):
        """
        Temporary build a solid from the user info, in order to retrieve information (volume etc).
        Can be used *before* initialization
        """
        vol = gate.new_element(user_info, self.simulation)
        vol = vol.build_solid()
        r = Box()
        r.cubic_volume = vol.GetCubicVolume()
        r.surface_area = vol.GetSurfaceArea()
        pMin = g4.G4ThreeVector()
        pMax = g4.G4ThreeVector()
        vol.BoundingLimits(pMin, pMax)
        r.bounding_limits = [pMin, pMax]
        return r

    def get_volume_depth(self, volume_name):
        depth = 0
        current = self.get_volume_user_info(volume_name)
        while current.name != "world":
            current = self.get_volume_user_info(current.mother)
            depth += 1
        return depth

    def _pop_keys_unused_by_solid(user_info):
        # remove unused keys: object, etc (it's a solid, not a volume)
        u = user_info.__dict__
        u.pop("mother", None)
        u.pop("translation", None)
        u.pop("color", None)
        u.pop("rotation", None)
        u.pop("material", None)

    def add_volume(self, vol_type, name):
        # check that another element with the same name does not already exist
        gate.assert_unique_element_name(self.user_info_volumes, name)
        # initialize the user_info
        v = gate.UserInfo("Volume", vol_type, name)
        # add to the list
        self.user_info_volumes[name] = v
        # FIXME  NOT CLEAR --> here ? or later
        # create a region for the physics cuts
        # user will be able to set stuff like :
        # pm.production_cuts.my_volume.gamma = 1 * mm
        pm = self.simulation.get_physics_user_info()
        cuts = pm.production_cuts
        cuts[name] = Box()
        # return the info
        return v

    def add_volume_from_solid(self, solid, name):
        v = None
        for op in gate.bool_operators:
            try:
                if op in solid:
                    v = self.add_volume("Boolean", name)
                    v.solid = solid
            except:
                pass
        if not v:
            v = self.add_volume(solid.type_name, name)
            # copy the parameters of the solid
            gate.copy_user_info(solid, v)
        return v

    def add_material_database(self, filename):
        if filename in self.material_database.filenames:
            gate.fatal(f'Database "{filename}" already exist.')
        self.material_database.read_from_file(filename)

    def dump_volumes(self):
        s = f"Number of volumes: {len(self.user_info_volumes)}"
        for vol in self.user_info_volumes.values():
            s += gate.indent(2, f"\n{vol}")
        return s

    def dump_tree_of_volumes(self):
        tree = gate.build_tree(self.simulation)
        info = {}
        for v in self.user_info_volumes.values():
            info[v.name] = v
        return gate.render_tree(tree, info)

    def _add_volume_to_tree(self, already_done, tree, vol):
        # check if mother volume exists
        if vol.mother not in self.user_info_volumes:
            gate.fatal(
                f"Cannot find a mother volume named '{vol.mother}', for the volume {vol}"
            )

        already_done[vol.name] = "in_progress"
        m = self.user_info_volumes[vol.mother]

        # check for the cycle
        if m.name not in already_done:
            self._add_volume_to_tree(already_done, tree, m)
        else:
            if already_done[m.name] == "in_progress":
                s = f"Error while building the tree, there is a cycle ? "
                s += f"\n volume is {vol}"
                s += f"\n parent is {m}"
                gate.fatal(s)

        # get the mother branch
        p = tree[m.name]

        # check not already exist
        if vol.name in tree:
            s = f"Node already exist in tree {vol.name} -> {tree}"
            s = s + f"\n Probably two volumes with the same name ?"
            gate.fatal(s)

        # create the node
        n = Node(vol.name, parent=p)
        tree[vol.name] = n
        already_done[vol.name] = True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation

paths = gate.get_default_test_paths(__file__, "gate_test029_volume_time_rotation")

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.visu = False
ui.random_seed = 123456

# units
m = gate.g4_units("m")
mm = gate.g4_units("mm")
cm = gate.g4_units("cm")
um = gate.g4_units("um")
nm = gate.g4_units("nm")
MeV = gate.g4_units("MeV")
Bq = gate.g4_units("Bq")
sec = gate.g4_units("second")

#  change world size
world = sim.world
world.size = [1 * m, 1 * m, 1 * m]

# add a simple fake volume to test hierarchy
# translation and rotation like in the Gate macro
fake = sim.add_volume("Box", "fake")
fake.size = [40 * cm, 40 * cm, 40 * cm]
fake.translation = [1 * cm, 2 * cm, 3 * cm]
fake.material = "G4_AIR"
fake.color = [1, 0, 1, 1]

# waterbox
waterbox = sim.add_volume("Box", "waterbox")
waterbox.mother = "fake"
waterbox.size = [20 * cm, 20 * cm, 20 * cm]
waterbox.translation = [-3 * cm, -2 * cm, -1 * cm]
waterbox.rotation = Rotation.from_euler("y", -20, degrees=True).as_matrix()
waterbox.material = "G4_WATER"
waterbox.color = [0, 0, 1, 1]

# physics
sim.set_cut("world", "all", 700 * um)

# default source for tests
# the source is fixed at the center, only the volume will move
source = sim.add_source("GenericSource", "mysource")
source.energy.mono = 150 * MeV
source.particle = "proton"
source.position.type = "disc"
source.position.radius = 5 * mm
source.direction.type = "momentum"
source.direction.momentum = [0, 0, 1]
source.activity = 20000 * Bq

# add dose actor
dose = sim.add_actor("DoseActor", "dose")
dose.output = paths.output / "test030-edep.mhd"
dose.mother = "waterbox"
dose.size = [99, 99, 99]
mm = gate.g4_units("mm")
dose.spacing = [2 * mm, 2 * mm, 2 * mm]
dose.translation = [2 * mm, 3 * mm, -2 * mm]
dose.uncertainty = True

# add stat actor
s = sim.add_actor("SimulationStatisticsActor", "Stats")
s.track_types_flag = True
s.output = paths.output / "stats030.txt"

# motion
motion = sim.add_actor("MotionVolumeActor", "Orbiting")
motion.mother = fake.name
motion.translations = []
motion.rotations = []
sim.run_timing_intervals = []
n = 3
start = 0
gantry_rotation = start
end = 1 * sec / n
for r in range(n):
    t, rot = gate.get_transform_orbiting(fake.translation, "Y", gantry_rotation)
    rot = gate.rot_np_as_g4(rot)
    motion.translations.append(t)
    motion.rotations.append(rot)
    sim.run_timing_intervals.append([start, end])
    gantry_rotation += 20
    start = end
    end += 1 * sec / n

# start simulation
output = sim.start()

# print results at the end
stat = output.get_actor("Stats")
print(stat)

dose = output.get_actor("dose")
print(dose)

# tests
stats_ref = gate.read_stat_file(paths.output_ref / "stats030.txt")
is_ok = gate.assert_stats(stat, stats_ref, 0.11)

print()
gate.warning("Difference for EDEP")
is_ok = (
    gate.assert_images(
        paths.output_ref / "test030-edep.mhd",
        paths.output / "test030-edep.mhd",
        stat,
        tolerance=30,
        ignore_value=0,
    )
    and is_ok
)

print("\nDifference for uncertainty")
is_ok = (
    gate.assert_images(
        paths.output_ref / "test030-edep_uncertainty.mhd",
        paths.output / "test030-edep_uncertainty.mhd",
        stat,
        tolerance=10,
        ignore_value=1,
    )
    and is_ok
)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pathlib

import opengate as gate


def dose_rate(param):
    # create the simulation
    sim = gate.Simulation()

    # main options
    ui = sim.user_info
    ui.g4_verbose = False
    ui.visu = param.visu
    ui.number_of_threads = param.number_of_threads
    ui.verbose_level = gate.INFO

    param.output_folder = pathlib.Path(param.output_folder)

    # units
    m = gate.g4_units("m")
    mm = gate.g4_units("mm")
    keV = gate.g4_units("keV")
    Bq = gate.g4_units("Bq")
    gcm3 = gate.g4_units("g/cm3")

    #  change world size
    world = sim.world
    world.size = [2 * m, 2 * m, 2 * m]

    # CT image
    ct = sim.add_volume("Image", "ct")
    ct.image = param.ct_image
    ct.material = "G4_AIR"  # material used by default
    tol = param.density_tolerance_gcm3 * gcm3
    ct.voxel_materials, materials = gate.HounsfieldUnit_to_material(
        tol, param.table_mat, param.table_density
    )
    if param.verbose:
        print(f'Density tolerance = {gate.g4_best_unit(tol, "Volumic Mass")}')
        print(f"Number of materials in the CT : {len(ct.voxel_materials)} materials")
    ct.dump_label_image = param.output_folder / "labels.mhd"

    # some radionuclides choice
    # (user of this function can still change
    # the source in the output sim)
    rad_list = {
        "Lu177": {"Z": 71, "A": 177, "name": "Lutetium 177"},
        "Y90": {"Z": 39, "A": 90, "name": "Yttrium 90"},
        "In111": {"Z": 49, "A": 111, "name": "Indium 111"},
        "I131": {"Z": 53, "A": 131, "name": "Iodine 131"},
    }

    # Activity source from an image
    source = sim.add_source("VoxelsSource", "vox")
    source.mother = ct.name
    source.particle = "ion"
    source.ion.Z = rad_list[param.radionuclide]["Z"]
    source.ion.A = rad_list[param.radionuclide]["A"]
    source.activity = param.activity_bq * Bq / ui.number_of_threads
    source.image = param.activity_image
    source.direction.type = "iso"
    source.energy.mono = 0 * keV
    # compute the translation to align the source with CT
    # (considering they are in the same physical space)
    source.position.translation = gate.get_translation_between_images_center(
        param.ct_image, param.activity_image
    )

    # cuts
    p = sim.get_physics_user_info()
    p.physics_list_name = "G4EmStandardPhysics_option4"
    p.enable_decay = True  # FIXME
    sim.set_cut("world", "all", 1 * m)
    sim.set_cut("ct", "all", 1 * mm)

    # add dose actor (get the same size as the source)
    source_info = gate.read_image_info(param.activity_image)
    dose = sim.add_actor("DoseActor", "dose")
    dose.output = param.output_folder / "edep.mhd"
    dose.mother = ct.name
    dose.size = source_info.size
    dose.spacing = source_info.spacing
    # translate the dose the same way as the source
    dose.translation = source.position.translation
    # set the origin of the dose like the source
    dose.img_coord_system = True
    dose.hit_type = "random"
    dose.uncertainty = False

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    return sim



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test028_ge_nm670_spect_4_acc_angle_helpers import *

# create the simulation
sim = gate.Simulation()

# simu description
spect, proj = create_spect_simu(
    sim,
    paths,
    number_of_threads=1,
    activity_kBq=1000,
    aa_enabled=False,
)

# go
output = sim.start()

# check
is_ok = compare_result(output, proj, "test028_aa_noaa.png")
gate.test_ok(is_ok)



from .ExceptionHandler import *


class Simulation:
    """
    Main class that store a simulation.
    It contains:
    - a set of global user parameters (SimulationUserInfo)
    - user parameters for Volume, Source, Actors and Filters, Physics
    - a list of g4 commands that will be set to G4 engine after the initialization

    There is NO Geant4 engine here, it is only a set of parameters and options.

    """

    def __init__(self, name="simulation"):
        """
        Constructor. Main members are:
        - managers of volumes, sources and actors
        - Geant4 objects that will be build during initialisation (start with g4_)
        - some internal variables
        """
        self.name = name

        # user's defined parameters
        self.user_info = gate.SimulationUserInfo(self)
        self.run_timing_intervals = None

        # list of G4 commands that will be called after
        # initialization and before start
        self.g4_commands = []

        # main managers
        self.volume_manager = gate.VolumeManager(self)
        self.source_manager = gate.SourceManager(self)
        self.actor_manager = gate.ActorManager(self)
        self.physics_manager = gate.PhysicsManager(self)
        self.filter_manager = gate.FilterManager(self)

        # default elements
        self._default_parameters()

    def __del__(self):
        pass

    def __str__(self):
        s = (
            f"Simulation name: {self.name} \n"
            f"Geometry       : {self.volume_manager}\n"
            f"Physics        : {self.physics_manager}\n"
            f"Sources        : {self.source_manager}\n"
            f"Actors         : {self.actor_manager}"
        )
        return s

    def _default_parameters(self):
        """
        Internal use.
        Build default elements: verbose, World, seed, physics, etc.
        """
        # World volume
        w = self.add_volume("Box", gate.__world_name__)
        w.mother = None
        m = gate.g4_units("meter")
        w.size = [3 * m, 3 * m, 3 * m]
        w.material = "G4_AIR"
        # run timing
        sec = gate.g4_units("second")
        self.run_timing_intervals = [
            [0 * sec, 1 * sec]
        ]  # a list of begin-end time values

    def dump_sources(self):
        return self.source_manager.dump()

    def dump_source_types(self):
        s = f""
        for t in gate.source_builders:
            s += f"{t} "
        return s

    def dump_volumes(self):
        return self.volume_manager.dump_volumes()

    def dump_tree_of_volumes(self):
        return self.volume_manager.dump_tree_of_volumes()

    def dump_volume_types(self):
        s = f""
        for t in gate.volume_builders:
            s += f"{t} "
        return s

    def dump_actors(self):
        return self.actor_manager.dump()

    def dump_actor_types(self):
        s = f""
        for t in gate.actor_builders:
            s += f"{t} "
        return s

    def dump_material_database_names(self):
        return list(self.volume_manager.material_database.filenames)

    def apply_g4_command(self, command):
        """
        For the moment, only use it *after* runManager.Initialize
        """
        self.g4_commands.append(command)

    @property
    def world(self):
        return self.get_volume_user_info(gate.__world_name__)

    def get_volume_user_info(self, name):
        v = self.volume_manager.get_volume_user_info(name)
        return v

    def get_all_volumes_user_info(self):
        return self.volume_manager.user_info_volumes

    def get_solid_info(self, user_info):
        return self.volume_manager.get_solid_info(user_info)

    def get_source_user_info(self, name):
        return self.source_manager.get_source_info(name)

    def get_actor_user_info(self, name):
        s = self.actor_manager.get_actor_user_info(name)
        return s

    def get_physics_user_info(self):
        return self.physics_manager.user_info

    def set_cut(self, volume_name, particle, value):
        self.physics_manager.set_cut(volume_name, particle, value)

    def set_physics_list(self, pl):
        p = self.get_physics_user_info()
        p.physics_list_name = pl

    def new_solid(self, solid_type, name):
        return self.volume_manager.new_solid(solid_type, name)

    def add_volume(self, solid_type, name):
        return self.volume_manager.add_volume(solid_type, name)

    def add_volume_from_solid(self, solid, name):
        return self.volume_manager.add_volume_from_solid(solid, name)

    def add_source(self, source_type, name):
        return self.source_manager.add_source(source_type, name)

    def add_actor(self, actor_type, name):
        return self.actor_manager.add_actor(actor_type, name)

    def add_filter(self, filter_type, name):
        return self.filter_manager.add_filter(filter_type, name)

    def add_material_database(self, filename):
        self.volume_manager.add_material_database(filename)

    def check_geometry(self):
        names = {}
        volumes = self.volume_manager.user_info_volumes
        for v in volumes:
            vol = volumes[v]

            # volume must have a name
            if "_name" not in vol.__dict__:
                gate.fatal(f"Volume is missing a 'name' : {vol}")

            # volume name must be geometry name
            if v != vol.name:
                gate.fatal(
                    f"Volume named '{v}' in geometry has a different name : {vol}"
                )

            if vol.name in names:
                gate.fatal(f"Two volumes have the same name '{vol.name}' --> {self}")
            names[vol.name] = True

            # volume must have a mother
            if "mother" not in vol.__dict__:
                gate.fatal(f"Volume is missing a 'mother' : {vol}")

            # volume must have a material
            if "material" not in vol.__dict__:
                gate.fatal(f"Volume is missing a 'material' : {vol}")

    def initialize(self):
        # self.current_engine = gate.SimulationEngine(self, start_new_process=False)
        gate.warning(f"(initialization do nothing)")

    def start(self, start_new_process=False):
        se = gate.SimulationEngine(self, start_new_process=start_new_process)
        return se.start()



import opengate as gate
import opengate_core as g4
import math


class SphereVolume(gate.VolumeBase):
    type_name = "Sphere"

    @staticmethod
    def set_default_user_info(user_info):
        gate.VolumeBase.set_default_user_info(user_info)
        user_info.rmin = 0
        user_info.rmax = 1
        user_info.sphi = 0
        user_info.dphi = 2 * math.pi
        user_info.stheta = 0
        user_info.dtheta = math.pi

    def build_solid(self):
        u = self.user_info
        return g4.G4Sphere(u.name, u.rmin, u.rmax, u.sphi, u.dphi, u.stheta, u.dtheta)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from test049_pet_digit_blurring_helpers import *
import opengate.contrib.spect_ge_nm670 as gate_spect
import matplotlib.pyplot as plt

paths = gate.get_default_test_paths(__file__, "")

"""
Check the options in DigitizerAdderActor
- TimeDifference: to store the max time difference within all hits in one single
- NumberOfHits: store the number of hits in one single
"""

# create the simulation
sim = gate.Simulation()

# units
m = gate.g4_units("m")
nm = gate.g4_units("nm")
mm = gate.g4_units("mm")
cm = gate.g4_units("cm")
Bq = gate.g4_units("Bq")
sec = gate.g4_units("s")
min = gate.g4_units("min")

# verbose
ui = sim.user_info
ui.visu = False
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.number_of_threads = 1
ui.random_seed = 123456

# world size
world = sim.world
world.size = [2.5 * m, 2.5 * m, 2.5 * m]
world.material = "G4_AIR"

# spect head without collimator
spect, crystal = gate_spect.add_ge_nm67_spect_head(
    sim, "spect", collimator_type=False, debug=False
)
spect.translation = [0, 0, -15 * cm]

# spect digitizer
hc = sim.add_actor("DigitizerHitsCollectionActor", f"Hits")
hc.mother = crystal.name
hc.output = ""  # No output
hc.attributes = [
    "PostPosition",
    "TotalEnergyDeposit",
    "PreStepUniqueVolumeID",
    "GlobalTime",
]
sc = sim.add_actor("DigitizerAdderActor", f"Singles")
sc.mother = hc.mother
sc.input_digi_collection = hc.name
sc.time_difference = True
sc.number_of_hits = True
sc.output = paths.output / "test051_singles.root"

# physics
p = sim.get_physics_user_info()
p.physics_list_name = "G4EmStandardPhysics_option3"
p.enable_decay = True
sim.set_cut("world", "all", 100 * mm)
sim.set_cut("spect", "all", 0.1 * mm)

# source of Ac225 ion
s1 = sim.add_source("GenericSource", "source")
s1.particle = "ion 89 225"  # Ac225
s1.position.type = "sphere"
s1.position.radius = 1 * nm
s1.position.translation = [0, 0, 0]
s1.direction.type = "iso"
s1.activity = 1e4 * Bq / ui.number_of_threads

# add stat actor
s = sim.add_actor("SimulationStatisticsActor", "stats")
s.track_types_flag = True
s.output = paths.output / "test051_stats.txt"

# go
output = sim.start(start_new_process=True)

# check stats
print()
gate.warning("Check stats")
stats = output.get_actor("stats")
stats_ref = gate.read_stat_file(paths.output_ref / "test051_stats.txt")
is_ok = gate.assert_stats(stats, stats_ref, tolerance=0.04)

# check root
print()
gate.warning("Check root time difference")
root1, n1 = gate.open_root_as_np(paths.output_ref / "test051_singles.root", "Singles")
root2, n2 = gate.open_root_as_np(sc.output, "Singles")

# time difference
td1 = root1["TimeDifference"]
td2 = root2["TimeDifference"]
td1 = td1[td1 > 1 * sec] / min
td2 = td2[td2 > 1 * sec] / min
is_ok = (
    gate.check_diff_abs(
        len(td1),
        len(td2),
        20,
        f"Number of time diff larger than 1 sec (~{len(td1)/n1*100:.2f}%):",
    )
    and is_ok
)
is_ok = (
    gate.check_diff_abs(np.mean(td1), np.mean(td2), 30, f"Time diff mean in minutes:")
    and is_ok
)

print()
gate.warning("Check root nb of hits")
nh1 = root1["NumberOfHits"]
nh2 = root2["NumberOfHits"]
is_ok = (
    gate.check_diff(np.mean(nh1), np.mean(nh2), 6, f"Number of hits in mean:") and is_ok
)

# plot
f, ax = plt.subplots(1, 2, figsize=(25, 10))
gate.plot_hist(ax[0], td1, f"Time diff ref (minutes)")
gate.plot_hist(ax[0], td2, f"Time diff (minutes)")
gate.plot_hist(ax[1], nh1, f"Nb of hits ref")
gate.plot_hist(ax[1], nh2, f"Nb of hits")

fn = paths.output / "test051_singles.png"
plt.savefig(fn)
print(f"Plot in {fn}")

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test033_rotation_spect_aa_helpers import *

paths = gate.get_default_test_paths(__file__, "")

# create the simulation
sim = gate.Simulation()
sources = create_test(sim, nb_thread=2)

# AA mode
for source in sources:
    source.direction.acceptance_angle.skip_policy = "SkipEvents"

# go
output = sim.start()

# check
is_ok = evaluate_test(output, sources, 13, 5913608)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test049_pet_digit_blurring_helpers import *
import numpy as np

paths = gate.get_default_test_paths(__file__, "gate_test049_pet_blur")

"""
see https://github.com/teaghan/PET_MonteCarlo
and https://doi.org/10.1002/mp.16032

PET simulation to test blurring options of the digitizer

- PET:
- phantom: nema necr
- output: singles with and without various blur options
"""

# create the simulation
sim = gate.Simulation()
create_simulation(sim, singles_name="Singles_readout")

# const
ns = gate.g4_units("ns")
keV = gate.g4_units("keV")
MeV = gate.g4_units("MeV")
sigma_to_fwhm = 2 * np.sqrt(2 * np.log(2))
fwhm_to_sigma = 1.0 / sigma_to_fwhm

# add (fake) blur
ro = sim.get_actor_user_info("Singles_readout")
bc1 = sim.add_actor("DigitizerBlurringActor", "Singles_1")
bc1.output = ro.output
bc1.input_digi_collection = "Singles_readout"
bc1.blur_attribute = "GlobalTime"
bc1.blur_method = "Gaussian"
bc1.blur_fwhm = 100 * ns

bc2 = sim.add_actor("DigitizerBlurringActor", "Singles")
bc2.output = ro.output
bc2.input_digi_collection = bc1.name
bc2.blur_attribute = "TotalEnergyDeposit"
bc2.blur_method = "Linear"
bc2.blur_resolution = 0.15
bc2.blur_reference_value = 511 * keV
bc2.blur_slope = -0.055 * (1 / MeV)

# start simulation
output = sim.start()

# print results
stats = output.get_actor("Stats")
print(stats)

# ----------------------------------------------------------------------------------------------------------
readout = output.get_actor("Singles_readout")
ig = readout.GetIgnoredHitsCount()
print()
print(f"Nb of ignored hits : {ig}")

# check stats
print()
gate.warning(f"Check stats")
p = paths.gate_output
stats_ref = gate.read_stat_file(p / "stats_blur2.txt")
is_ok = gate.assert_stats(stats, stats_ref, 0.03)

# check root singles
f = p / "pet_blur2.root"
bc = output.get_actor("Singles").user_info
is_ok = (
    check_root_singles(paths, 1, f, bc.output, png_output="test049_singles_wb2.png")
    and is_ok
)

# timing
b = check_timing(f, bc.output)
is_ok = is_ok and b

gate.test_ok(is_ok)



class EngineBase:
    """
    Base class for all engines (SimulationEngine, VolumeEngine, etc.)
    """

    def __init__(self):
        # debug verbose
        self.verbose_destructor = False

    def __del__(self):
        if self.verbose_destructor:
            print("del EngineBase")



import opengate as gate
import opengate_core as g4
from box import Box


class PhysicsManager:
    """
    Everything related to the physics (lists, cuts etc) should be here.
    """

    def __init__(self, simulation):
        # Keep a pointer to the current simulation
        self.simulation = simulation
        # user options
        self.user_info = gate.PhysicsUserInfo(self)
        # default values
        self._default_parameters()
        # names for particle cuts
        self.cut_particle_names = {
            "gamma": "gamma",
            "electron": "e-",
            "positron": "e+",
            "proton": "proton",
        }

    def __del__(self):
        # not really clear, but it seems that we should delete user_info here
        # if not seg fault (sometimes) at the end
        # print("del PhysicsManager (then del user info)")
        del self.user_info
        pass

    def __str__(self):
        s = f"{self.user_info.physics_list_name} Decay: {self.user_info.enable_decay}"
        return s

    def _default_parameters(self):
        ui = self.user_info
        # keep the name to be able to come back to default
        self.default_physic_list = "QGSP_BERT_EMV"
        ui.physics_list_name = self.default_physic_list
        ui.enable_decay = False
        ui.production_cuts.world = Box()
        ui.production_cuts.world.gamma = -1  # -1 means = will be the phys list default
        ui.production_cuts.world.proton = -1
        ui.production_cuts.world.electron = -1
        ui.production_cuts.world.positron = -1
        ui.production_cuts.world.propagate_to_daughters = True
        """
        FIXME Energy range not clear : does not work in mono-thread mode
        Ignored for the moment (keep them to None)
        """
        """
        eV = gate.g4_units('eV')
        keV = gate.g4_units('keV')
        GeV = gate.g4_units('GeV')
        ui.energy_range_min = 250 * eV
        ui.energy_range_max = 0.5 * GeV
        """
        ui.energy_range_min = None
        ui.energy_range_max = None
        ui.apply_cuts = True

    def dump_available_physics_lists(self):
        factory = g4.G4PhysListFactory()
        s = (
            f"Phys List:     {factory.AvailablePhysLists()}\n"
            f"Phys List EM:  {factory.AvailablePhysListsEM()}\n"
            f"Phys List add: {gate.available_additional_physics_lists}"
        )
        return s

    def dump_cuts(self):
        s = ""
        for c in self.user_info.production_cuts:
            s += f"{c} : {self.user_info.production_cuts[c]}\n"
        return s

    def set_cut(self, volume_name, particle_name, value):
        cuts = self.user_info.production_cuts
        if volume_name not in cuts:
            s = f'Cannot find the volume "{volume_name}" to define its cut.'
            gate.fatal(s)
        if particle_name == "all":
            cuts[volume_name]["gamma"] = value
            cuts[volume_name]["electron"] = value
            cuts[volume_name]["positron"] = value
            cuts[volume_name]["proton"] = value
            return
        if particle_name not in cuts[volume_name]:
            s = f'Cannot find the particle named "{particle_name}" to define its cut in the volume "{volume_name}".'
            gate.fatal(s)
        cuts[volume_name][particle_name] = value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from box import Box
from opengate.contrib.dose_rate_helpers import dose_rate

paths = gate.get_default_test_paths(__file__, "")
dr_data = paths.data / "dose_rate_data"

# set param
gcm3 = gate.g4_units("g/cm3")
param = Box()
param.ct_image = str(dr_data / "29_CT_5mm_crop.mhd")
param.table_mat = str(dr_data / "Schneider2000MaterialsTable.txt")
param.table_density = str(dr_data / "Schneider2000DensitiesTable.txt")
param.activity_image = str(dr_data / "activity_test_crop_4mm.mhd")
param.radionuclide = "Lu177"
param.activity_bq = 1e6
param.number_of_threads = 1
param.visu = False
param.verbose = True
param.density_tolerance_gcm3 = 0.05
param.output_folder = str(paths.output / "output_test035")

# Create the simu
# Note that the returned sim object can be modified to change source or cuts or whatever other parameters
sim = dose_rate(param)

# Change source to alpha to get quick high local dose
source = sim.get_source_user_info("vox")
source.particle = "alpha"
MeV = gate.g4_units("MeV")
source.energy.mono = 1 * MeV

print("Phys list cuts:")
print(sim.physics_manager.dump_cuts())

# run
output = sim.start(True)

# print results
print()
gate.warning(f"Check stats")
stats = output.get_actor("Stats")
stats.write(param.output_folder / "stats035.txt")
print(stats)
stats_ref = gate.read_stat_file(paths.output_ref / "output_test035" / "stats.txt")
is_ok = gate.assert_stats(stats, stats_ref, 0.10)

# dose comparison
print()
gate.warning(f"Check dose")
h = output.get_actor("dose")
print(h)
is_ok = (
    gate.assert_images(
        paths.output_ref / "output_test035" / "edep.mhd",
        h.user_info.output,
        stats,
        tolerance=15,
        ignore_value=0,
    )
    and is_ok
)

gate.test_ok(is_ok)



import opengate as gate
import opengate_core as g4


class DigitizerReadoutActor(g4.GateDigitizerReadoutActor, gate.ActorBase):
    """
    This actor is a DigitizerAdderActor + a discretization step:
    the final position is the center of the volume
    """

    type_name = "DigitizerReadoutActor"

    @staticmethod
    def set_default_user_info(user_info):
        gate.DigitizerAdderActor.set_default_user_info(user_info)
        user_info.discretize_volume = None

    def __init__(self, user_info):
        gate.ActorBase.__init__(self, user_info)
        g4.GateDigitizerReadoutActor.__init__(self, user_info.__dict__)
        actions = {"StartSimulationAction", "EndSimulationAction"}
        self.AddActions(actions)

    def __del__(self):
        pass

    def __str__(self):
        s = f"DigitizerReadoutActor {self.user_info.name}"
        return s

    def StartSimulationAction(self):
        gate.DigitizerAdderActor.set_group_by_depth(self)
        if self.user_info.discretize_volume is None:
            gate.fatal(f'Please, set the option "discretize_volume"')
        depth = self.simulation.volume_manager.get_volume_depth(
            self.user_info.discretize_volume
        )
        self.SetDiscretizeVolumeDepth(depth)
        g4.GateDigitizerReadoutActor.StartSimulationAction(self)

    def EndSimulationAction(self):
        g4.GateDigitizerReadoutActor.EndSimulationAction(self)



from .GenericSource import *
from .GANSourceDefaultGenerator import GANSourceDefaultGenerator
import sys


class GANSource(GenericSource):
    """
    GAN source: the Generator produces particles
    Input is a neural network Generator trained with a GAN
    """

    type_name = "GANSource"

    @staticmethod
    def set_default_user_info(user_info):
        GenericSource.set_default_user_info(user_info)
        # additional param
        user_info.pth_filename = None
        user_info.position_keys = None
        user_info.backward_distance = None
        # if backward is enabled and the time is not managed by the GAN,
        # the time cannot be changed (yet). Use 'force' to enable backward
        user_info.backward_force = False
        user_info.direction_keys = None
        user_info.energy_key = None
        user_info.energy_min_threshold = -1
        user_info.energy_max_threshold = sys.float_info.max
        user_info.weight_key = None
        user_info.time_key = None
        user_info.relative_timing = True
        user_info.batch_size = 10000
        user_info.generator = None
        user_info.verbose_generator = False
        user_info.use_time = False
        user_info.use_weight = False
        # specific to conditional GAN
        user_info.cond_image = None
        user_info.compute_directions = False
        user_info.cond_debug = False
        # for skipped particles
        user_info.skip_policy = "SkipEvents"  # or ZeroEnergy

    def __del__(self):
        pass

    def create_g4_source(self):
        return g4.GateGANSource()

    def __init__(self, user_info):
        super().__init__(user_info)

    def initialize(self, run_timing_intervals):
        # FIXME -> check input user_info

        # initialize the mother class generic source
        GenericSource.initialize(self, run_timing_intervals)

        # default generator or set by the user
        if self.user_info.generator is None:
            self.set_default_generator()
        gen = self.user_info.generator

        # initialize the generator (read the GAN)
        # this function must have 1) the generator function 2) the associated info
        gen.initialize()

        # set the function pointer to the cpp side
        self.g4_source.SetGeneratorFunction(gen.generator)

        # set the parameters to the cpp side
        self.g4_source.SetGeneratorInfo(gen.gan_info)

    def set_default_generator(self):
        # non-conditional generator
        if self.user_info.cond_image is None:
            self.user_info.generator = GANSourceDefaultGenerator(self.user_info)
            return

        vcg = gate.VoxelizedSourceConditionGenerator(self.user_info.cond_image, self)
        vcg.compute_directions = self.user_info.compute_directions
        g = gate.GANSourceConditionalGenerator(self.user_info, vcg.generate_condition)
        self.user_info.generator = g



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from test013_phys_lists_helpers import create_pl_sim, phys_em_parameters

paths = gate.get_default_test_paths(__file__, "gate_test013_phys_lists")

# create simulation
sim = create_pl_sim()
ui = sim.user_info
ui.g4_verbose = True

# remove ion sources
sim.source_manager.user_info_sources.pop("ion1")
sim.source_manager.user_info_sources.pop("ion2")

# change physics
p = sim.get_physics_user_info()
p.physics_list_name = "G4EmStandardPhysics_option4"
p.enable_decay = True
p.apply_cuts = True  # default
cuts = p.production_cuts
um = gate.g4_units("um")
cuts.world.gamma = 7 * um
cuts.world.electron = 7 * um
cuts.world.positron = 7 * um
cuts.world.proton = 7 * um

# em parameters
# phys_em_parameters(p)

# print cuts
print("Phys list cuts:")
print(sim.physics_manager.dump_cuts())

# start simulation
# sim.apply_g4_command("/tracking/verbose 1")
output = sim.start()

# Gate mac/main_1.mac
stats = output.get_actor("Stats")
stats_ref = gate.read_stat_file(paths.gate_output / "stat_1.txt")
is_ok = gate.assert_stats(stats, stats_ref, tolerance=0.1)

gate.test_ok(is_ok)



from .GANSource import *
from .GANSourceDefaultPairsGenerator import GANSourceDefaultPairsGenerator


class GANPairsSource(GANSource):
    """
    GAN source: the Generator produces pairs of particles (for PET)
    Input is a neural network Generator trained with a GAN
    """

    type_name = "GANPairsSource"

    @staticmethod
    def set_default_user_info(user_info):
        GANSource.set_default_user_info(user_info)

    def __del__(self):
        pass

    def create_g4_source(self):
        return g4.GateGANPairSource()

    def __init__(self, user_info):
        super().__init__(user_info)

    def set_default_generator(self):
        # non-conditional generator
        if self.user_info.cond_image is None:
            self.user_info.generator = GANSourceDefaultPairsGenerator(self.user_info)
            return

        # conditional generator
        gate.fatal(
            f"A conditional generator must be set in the "
            f"user_info.generator option of the GANPairsSource '{self.user_info.name}'."
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import click
from test045_gan_phsp_pet_gan_helpers import *

paths = gate.get_default_test_paths(__file__, "")

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option("--phantom", "-p", default="analytic", help="Phantom : analytic or vox")
@click.option("--source", "-s", default="analytic", help="Source : analytic or vox")
@click.option("--rad", "-r", default="Ga68", help="Radionuclide Ga68 or F18, etc")
@click.option("--gaga", default=False, is_flag=True, help="Use gaga (GAN) or not")
@click.option("--pet", default=False, is_flag=True, help="W/wo PET output")
@click.option("--activity_bqml", "-a", default="10", help="Activity in BqmL")
@click.option("--threads", "-t", default=1, help="Nb of threads")
@click.option("--visu", "-v", default=False, is_flag=True, help="visu for debug")
@click.option(
    "--output_folder", "-o", default=".", help="output folder (AUTO for the tests)"
)
@click.option("--seed", default="auto", help="random engine seed")
def go(
    phantom, source, rad, gaga, pet, activity_bqml, visu, threads, output_folder, seed
):
    run_test_045_speedrun(
        phantom,
        source,
        rad,
        gaga,
        pet,
        activity_bqml,
        visu,
        threads,
        output_folder,
        seed,
    )


def run_test_045_speedrun(
    phantom, source, rad, gaga, pet, activity_bqml, visu, threads, output_folder, seed
):
    # WARNING
    # only to check computation times, I did not check results

    p = Box()
    p.phantom_type = phantom
    p.source_type = source
    p.use_gaga = gaga
    p.use_pet = pet
    p.activity_Bqml = float(activity_bqml)
    p.number_of_threads = threads
    p.radionuclide = rad

    # (could be options)
    p.iec_vox_mhd = str(paths.data / "iec_2mm.mhd")
    p.iec_vox_json = paths.data / "iec_2mm.json"
    p.source_vox_mhd = str(paths.data / "iec_source_4mm.mhd")
    p.gaga_pth = paths.data / "pth120_test9221_GP_0GP_10.0_100000.pth"

    gate.print_dic(p)

    # output
    if output_folder == "AUTO":
        output_folder = str(paths.output)
    out = f"test045_speedup_p_{p.phantom_type}_s_{p.source_type}_pet_{p.use_pet}_gaga_{gaga}"
    p.pet_output = f"{output_folder}/{out}.root"

    # init the simulation
    sim = gate.Simulation()

    # visu
    sim.user_info.visu = visu
    if visu:
        p.iec_vox_mhd = paths.data / "5x5x5.mhd"
        p.activity_Bqml = 1

    # seed
    if seed != "auto":
        seed = int(seed)
    sim.user_info.random_seed = seed

    # create the simulation
    create_pet_simulation(sim, p)

    # warning cuts

    output = sim.start()

    # print
    stats = output.get_actor("Stats")
    print(stats)

    # save
    stats.write(f"{output_folder}/{out}.txt")

    if p.use_pet:
        import uproot

        phsp = output.get_actor("Singles").user_info
        f = phsp.output
        s = uproot.open(f)["Singles"]
        print(f"Number of singles: ", s.num_entries)

    return stats.pps


# --------------------------------------------------------------------------
if __name__ == "__main__":
    go()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate

paths = gate.get_default_test_paths(__file__, "gate_test034_gan_phsp_linac")

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.visu = False
ui.check_volumes_overlap = False
ui.number_of_threads = 1
# ui.running_verbose_level = gate.EVENT

# units
m = gate.g4_units("m")
mm = gate.g4_units("mm")
cm = gate.g4_units("cm")
nm = gate.g4_units("nm")
Bq = gate.g4_units("Bq")
kBq = 1000 * Bq
MBq = 1000 * kBq
MeV = gate.g4_units("MeV")

#  adapt world size
world = sim.world
world.size = [2 * m, 2 * m, 2 * m]
world.material = "G4_AIR"

# FIXME to compare to current GATE benchmark gaga
# FIXME or the dqprm exercises write/read

# add a waterbox
waterbox = sim.add_volume("Box", "waterbox")
waterbox.size = [30 * cm, 30 * cm, 30 * cm]
waterbox.translation = [0 * cm, 0 * cm, 52.2 * cm]
waterbox.material = "G4_WATER"
waterbox.color = [0, 0, 1, 1]  # blue

# virtual plane for phase space
# It is not really used, only for visualisation purpose
# and as origin of the coordinate system of the GAN source
plane = sim.add_volume("Box", "phase_space_plane")
plane.mother = world.name
plane.material = "G4_AIR"
plane.size = [3 * cm, 4 * cm, 5 * cm]
# plane.rotation = Rotation.from_euler('x', 15, degrees=True).as_matrix()
plane.color = [1, 0, 0, 1]  # red

# GAN source
# in the GAN : position, direction, E, weights
gsource = sim.add_source("GANSource", "gaga")
gsource.particle = "gamma"
gsource.mother = plane.name
# gsource.activity = 10 * MBq / ui.number_of_threads
gsource.n = 1e6 / ui.number_of_threads
gsource.pth_filename = (
    paths.data / "003_v3_40k.pth"
)  # FIXME also allow .pt (include the NN)
gsource.position_keys = ["X", "Y", 271.1 * mm]
gsource.direction_keys = ["dX", "dY", "dZ"]
gsource.energy_key = "Ekine"
gsource.weight_key = None
gsource.time_key = None
gsource.batch_size = 1e5
gsource.verbose_generator = True
# it is possible to define another generator
# gsource.generator = generator

# add stat actor
s = sim.add_actor("SimulationStatisticsActor", "Stats")
s.track_types_flag = True

# PhaseSpace Actor
dose = sim.add_actor("DoseActor", "dose")
dose.mother = waterbox.name
dose.spacing = [4 * mm, 4 * mm, 4 * mm]
dose.size = [75, 75, 75]
dose.output = paths.output / "test034_edep.mhd"
dose.uncertainty = True

"""
Dont know why similar to hit_type == post while in Gate
this is hit_type = random ?
"""
dose.hit_type = "post"

# phys
p = sim.get_physics_user_info()
p.physics_list_name = "G4EmStandardPhysics_option4"
sim.set_cut("world", "all", 1000 * m)
sim.set_cut("waterbox", "all", 1 * mm)

# start simulation
output = sim.start()

s = output.get_source("gaga")
print(f"Source, nb of E<=0: {s.fTotalSkippedEvents}")

# print results
gate.warning(f"Check stats")
stats = output.get_actor("Stats")
print(stats)
stats_ref = gate.read_stat_file(paths.gate / "stats.txt")
is_ok = gate.assert_stats(stats, stats_ref, 0.10)

gate.warning(f"Check dose")
h = output.get_actor("dose")
print(h)
is_ok = (
    gate.assert_images(
        paths.gate / "dose-Edep.mhd", dose.output, stats, tolerance=58, ignore_value=0
    )
    and is_ok
)

gate.test_ok(is_ok)



import opengate as gate
import opengate_core as g4
import uuid
import time


class TestActor(g4.GateVActor, gate.ActorBase):
    """
    Test actor: only py side (no cpp)
    For prototyping (slow)
    """

    type_name = "TestActor"

    @staticmethod
    def set_default_user_info(user_info):
        gate.ActorBase.set_default_user_info(user_info)
        user_info.track_types_flag = False

    def __init__(self, user_info=None):
        # user_info can be null when create empty actor (that read file)
        if not user_info:
            user_info = gate.UserInfo(
                "Actor", self.type_name, name=uuid.uuid4().__str__()
            )
        gate.ActorBase.__init__(self, user_info)
        g4.GateVActor.__init__(self, user_info.__dict__)
        actions = {
            "StartSimulationAction",
            "EndSimulationAction",
            "BeginOfEventAction",
            "EndOfRunAction",
            "PreUserTrackingAction",
            "SteppingAction",
        }
        self.AddActions(actions)
        # empty results for the moment
        self.run_count = 0
        self.event_count = 0
        self.track_count = 0
        self.step_count = 0
        self.duration = 0
        self.track_types = {}
        self.start_time = 0
        self.end_time = 0

    def __del__(self):
        pass

    @property
    def pps(self):
        sec = gate.g4_units("s")
        if self.duration != 0:
            return self.event_count / self.duration * sec
        return 0

    @property
    def tps(self):
        sec = gate.g4_units("s")
        if self.duration != 0:
            return self.track_count / self.duration * sec
        return 0

    @property
    def sps(self):
        sec = gate.g4_units("s")
        if self.duration != 0:
            return self.step_count / self.duration * sec
        return 0

    def __str__(self):
        if not self:
            return ""
        s = (
            f"Runs     {self.run_count}\n"
            f"Events   {self.event_count}\n"
            f"Tracks   {self.track_count}\n"
            f"Step     {self.step_count}\n"
            f'Duration {g4.G4BestUnit(self.duration, "Time")}\n'
            f"PPS      {self.pps:.0f}\n"
            f"TPS      {self.tps:.0f}\n"
            f"SPS      {self.sps:.0f}"
        )
        if self.user_info.track_types_flag:
            s += f"\n" f"Track types: {self.track_types}"
        return s

    def StartSimulationAction(self):
        self.start_time = time.time()

    def BeginOfEventAction(self, event):
        pass

    def PreUserTrackingAction(self, track):
        self.track_count += 1
        if self.user_info.track_types_flag:
            p = track.GetParticleName()
            try:
                self.track_types[p] += 1
            except:
                self.track_types[p] = 1

    def EndOfRunAction(self, run):
        self.run_count += 1
        self.event_count += run.GetNumberOfEvent()

    def SteppingAction(self, step, touchable):
        self.step_count += 1

    def EndSimulationAction(self):
        self.end_time = time.time()
        sec = gate.g4_units("s")
        self.duration = (self.end_time - self.start_time) * sec

    def write(self, filename):
        sec = gate.g4_units("s")
        f = open(filename, "w+")
        s = f"# NumberOfRun    = {self.run_count}\n"
        s += f"# NumberOfEvents = {self.event_count}\n"
        s += f"# NumberOfTracks = {self.track_count}\n"
        s += f"# NumberOfSteps  = {self.step_count}\n"
        s += f"# NumberOfGeometricalSteps  = ?\n"
        s += f"# NumberOfPhysicalSteps     = ?\n"
        s += f"# ElapsedTime           = {self.duration / sec}\n"
        s += f"# ElapsedTimeWoInit     = {self.duration / sec}\n"
        s += f"# StartDate             = ?\n"
        s += f"# EndDate               = ?\n"
        s += f"# PPS (Primary per sec)      = {self.pps:.0f}\n"
        s += f"# TPS (Track per sec)        = {self.tps:.0f}\n"
        s += f"# SPS (Step per sec)         = {self.sps:.0f}\n"
        if self.user_info.track_types_flag:
            s += f"# Track types:\n"
            for t in self.track_types:
                s += f"# {t} = {self.track_types[t]}\n"
        f.write(s)



from opengate import log
import time
import random
import sys
import os
import opengate_core as g4
from .ExceptionHandler import *
from multiprocessing import Process, set_start_method, Queue


class SimulationEngine(gate.EngineBase):
    """
    Main class to execute a Simulation (optionally in a separate subProcess)
    """

    def __init__(self, simulation, start_new_process=False):
        gate.EngineBase.__init__(self)

        # current state of the engine
        self.state = "before"  # before | started | after
        self.is_initialized = False

        # do we create a subprocess or not ?
        self.start_new_process = start_new_process

        # LATER : option to wait the end of completion or not

        # store the simulation object
        self.simulation = simulation

        # UI
        self.ui_session = None
        self.g4_ui = None

        # all engines
        self.volume_engine = None
        self.physics_engine = None
        self.source_engine = None
        self.action_engine = None
        self.actor_engine = None

        # random engine
        self.g4_HepRandomEngine = None
        self.current_random_seed = None

        # Main Run Manager
        self.g4_RunManager = None

        # exception handler
        self.g4_exception_handler = None

        # user fct to call after initialization
        self.user_fct_after_init = None

    def __del__(self):
        if self.verbose_destructor:
            print("del SimulationEngine")

        # Set verbose to zero before destructor to avoid the final message
        # This is needed to avoid seg fault when run in a sub process
        if getattr(self, "g4_RunManager", False):
            self.g4_RunManager.SetVerboseLevel(0)
        pass

    def start(self):
        # set start method only work on linux and osx, not windows
        # https://superfastpython.com/multiprocessing-spawn-runtimeerror/
        # Alternative: put the
        # if __name__ == '__main__':
        # at the beginning of the script

        # Check when GDML is activated, if G4 was compiled with GDML
        if (
            self.simulation.user_info.visu == True
            and self.simulation.user_info.visu_type == "gdml"
        ):
            gi = g4.GateInfo
            if not gi.get_G4GDML():
                return

        if self.start_new_process:
            # https://britishgeologicalsurvey.github.io/science/python-forking-vs-spawn/
            # (the "force" option is needed for notebooks)
            set_start_method("fork", force=True)
            # set_start_method("spawn")
            q = Queue()
            p = Process(target=self.init_and_start, args=(q,))
            p.start()
            self.state = "started"
            p.join()
            self.state = "after"
            output = q.get()
        else:
            output = self.init_and_start(None)

        # put back the simulation object to all actors
        for actor in output.actors.values():
            actor.simulation = self.simulation
        output.simulation = self.simulation

        # start visualization if vrml or gdml
        if (
            self.simulation.user_info.visu == True
            and self.simulation.user_info.visu_type == "vrml"
        ):
            self.vrml_visualization()
        elif (
            self.simulation.user_info.visu == True
            and self.simulation.user_info.visu_type == "gdml"
        ):
            self.gdml_visualization()

        # return the output of the simulation
        return output

    def init_and_start(self, queue):
        self.state = "started"

        # go
        self.initialize()
        self.apply_all_g4_commands()
        if self.user_fct_after_init:
            log.info("Simulation: initialize user fct")
            self.user_fct_after_init(self)
        self._start()

        # prepare the output
        output = gate.SimulationOutput()
        output.store_actors(self)
        output.store_sources(self)
        output.current_random_seed = self.current_random_seed
        if queue is not None:
            queue.put(output)
            return None
        return output

    def initialize(self):
        """
        Build the main geant4 objects and initialize them.
        """
        # shorter code
        ui = self.simulation.user_info

        # g4 verbose
        self.initialize_g4_verbose()

        # check multithreading
        mt = g4.GateInfo.get_G4MULTITHREADED()
        if ui.number_of_threads > 1 and not mt:
            gate.fatal(
                "Cannot use multi-thread, opengate_core was not compiled with Geant4 MT"
            )

        # init random engine (before the MTRunManager creation)
        self.initialize_random_engine()

        # create the RunManager
        if ui.number_of_threads > 1 or ui.force_multithread_mode:
            log.info(
                f"Simulation: create MTRunManager with {ui.number_of_threads} threads"
            )
            rm = g4.G4RunManagerFactory.CreateMTRunManager(ui.number_of_threads)
            rm.SetNumberOfThreads(ui.number_of_threads)
        else:
            log.info("Simulation: create RunManager")
            rm = g4.G4RunManagerFactory.CreateRunManager()

        if rm is None:
            self.fatal("no RunManager")

        self.g4_RunManager = rm
        self.g4_RunManager.SetVerboseLevel(ui.g4_verbose_level)

        # create the handler for the exception
        self.g4_exception_handler = ExceptionHandler()

        # check run timing
        self.run_timing_intervals = self.simulation.run_timing_intervals.copy()
        gate.assert_run_timing(self.run_timing_intervals)

        # geometry
        log.info("Simulation: initialize Geometry")
        self.volume_engine = gate.VolumeEngine(self.simulation)
        self.volume_engine.verbose_destructor = self.verbose_destructor
        self.g4_RunManager.SetUserInitialization(self.volume_engine)

        # phys
        log.info("Simulation: initialize Physics")
        self.physics_engine = gate.PhysicsEngine(self.simulation.physics_manager)
        self.physics_engine.initialize()
        self.g4_RunManager.SetUserInitialization(self.physics_engine.g4_physic_list)

        # sources
        log.info("Simulation: initialize Source")
        self.source_engine = gate.SourceEngine(self.simulation.source_manager)
        self.source_engine.initialize(self.simulation.run_timing_intervals)

        # action
        log.info("Simulation: initialize Actions")
        self.action_engine = gate.ActionEngine(self.source_engine)
        self.g4_RunManager.SetUserInitialization(self.action_engine)

        # Actors initialization (before the RunManager Initialize)
        self.actor_engine = gate.ActorEngine(self.simulation.actor_manager, self)
        self.actor_engine.create_actors()
        self.source_engine.initialize_actors(self.actor_engine.actors)
        self.volume_engine.set_actor_engine(self.actor_engine)

        # Initialization
        log.info("Simulation: initialize G4RunManager")
        self.g4_RunManager.Initialize()
        self.is_initialized = True

        # Physics cuts initialization
        log.info("Simulation: initialize Physics cuts")
        tree = self.volume_engine.volumes_tree
        self.physics_engine.initialize_cuts(tree)

        # Actors initialization
        log.info("Simulation: initialize Actors")
        self.actor_engine.initialize()

        # Check overlaps
        if ui.check_volumes_overlap:
            log.info("Simulation: check volumes overlap")
            self.check_volumes_overlap(verbose=False)
        else:
            log.info("Simulation: (no volumes overlap checking)")

        # Register sensitive detector.
        # if G4 was compiled with MT (regardless it is used or not)
        # ConstructSDandField (in VolumeManager) will be automatically called
        if not g4.GateInfo.get_G4MULTITHREADED():
            gate.warning("DEBUG Register sensitive detector in no MT mode")
            self.actor_engine.register_sensitive_detectors()

        # vrml initialization
        if (
            self.simulation.user_info.visu == True
            and (ui.visu_type == "vrml_file_only" or ui.visu_type == "vrml")
            and ui.visu_filename
        ):
            os.environ["G4VRMLFILE_FILE_NAME"] = ui.visu_filename

        # gdml initialization
        if (
            self.simulation.user_info.visu == True
            and (ui.visu_type == "gdml_file_only" or ui.visu_type == "gdml")
            and ui.visu_filename
        ):
            if os.path.isfile(ui.visu_filename):
                os.remove(ui.visu_filename)

    def apply_all_g4_commands(self):
        n = len(self.simulation.g4_commands)
        if n > 0:
            log.info(f"Simulation: apply {n} G4 commands")
        for command in self.simulation.g4_commands:
            self.apply_g4_command(command)

    def gdml_visualization(self):
        try:
            import pyg4ometry
        except:
            print(
                "The module pyg4ometry is not installed to be able to visualize gdml files. Execute:"
            )
            print("pip install pyg4ometry")
            return
        r = pyg4ometry.gdml.Reader(self.simulation.user_info.visu_filename)
        l = r.getRegistry().getWorldVolume()
        v = pyg4ometry.visualisation.VtkViewerColouredMaterial()
        v.addLogicalVolume(l)
        v.view()

    def vrml_visualization(self):
        try:
            import pyvista
        except:
            print(
                "The module pyvista is not installed to be able to visualize vrml files. Execute:"
            )
            print("pip install pyvista")
            return
        pl = pyvista.Plotter()
        pl.import_vrml(self.simulation.user_info.visu_filename)
        pl.add_axes(line_width=5)
        pl.show()

    def apply_g4_command(self, command):
        if self.g4_ui is None:
            self.g4_ui = g4.G4UImanager.GetUIpointer()
        self.g4_ui.ApplyCommand(command)

    def _start(self):
        """
        Start the simulation. The runs are managed in the SourceManager.
        """
        s = ""
        if self.start_new_process:
            s = "(in a new process)"
        log.info("-" * 80 + f"\nSimulation: START {s}")

        # visualisation should be initialized *after* other initializations ?
        # FIXME self._initialize_visualisation()

        # actor: start simulation (only the master thread)
        self.actor_engine.start_simulation()

        # go !
        start = time.time()
        self.source_engine.start()
        end = time.time()

        # actor: stop simulation (only the master thread)
        self.actor_engine.stop_simulation()

        # this is the end
        log.info(
            f"Simulation: STOP. Run: {len(self.run_timing_intervals)}. "
            # f'Events: {self.source_manager.total_events_count}. '
            f"Time: {end - start:0.1f} seconds.\n"
            + f"-" * 80
        )

    def initialize_random_engine(self):
        engine_name = self.simulation.user_info.random_engine
        self.g4_HepRandomEngine = None
        if engine_name == "MixMaxRng":
            self.g4_HepRandomEngine = g4.MixMaxRng()
        if engine_name == "MersenneTwister":
            self.g4_HepRandomEngine = g4.MTwistEngine()
        if not self.g4_HepRandomEngine:
            s = f"Cannot find the random engine {engine_name}\n"
            s += f"Use: MersenneTwister or MixMaxRng"
            gate.fatal(s)

        # set the random engine
        g4.G4Random.setTheEngine(self.g4_HepRandomEngine)
        if self.simulation.user_info.random_seed == "auto":
            self.current_random_seed = random.randrange(sys.maxsize)
        else:
            self.current_random_seed = self.simulation.user_info.random_seed

        # set the seed
        g4.G4Random.setTheSeed(self.current_random_seed, 0)

    def initialize_g4_verbose(self):
        if not self.simulation.user_info.g4_verbose:
            # no Geant4 output
            ui = gate.UIsessionSilent()
        else:
            # Geant4 output with color
            ui = gate.UIsessionVerbose()
        # it is also possible to set ui=None for 'default' output
        # we must keep a ref to ui_session
        self.ui_session = ui
        # we must keep a ref to ui_manager
        self.g4_ui = g4.G4UImanager.GetUIpointer()
        if self.g4_ui is None:
            self.fatal("no g4_ui")
        self.g4_ui.SetCoutDestination(ui)

    def fatal(self, err=""):
        s = (
            f"Cannot run a new simulation in this process: only one execution is possible.\n"
            f"Use the option start_new_process=True in gate.SimulationEngine. {err}"
        )
        gate.fatal(s)

    def check_volumes_overlap(self, verbose=True):
        # FIXME: later, allow to bypass this check ?

        # we need to 'cheat' the verbosity before doing the check
        ui = self.simulation.user_info
        b = ui.g4_verbose
        ui.g4_verbose = True
        self.initialize_g4_verbose()

        # check
        self.volume_engine.check_overlaps(verbose)

        # put back verbosity
        ui.g4_verbose = b
        self.initialize_g4_verbose()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.spect_ge_nm670 as gate_spect
import opengate.contrib.phantom_nema_iec_body as gate_iec
from scipy.spatial.transform import Rotation

paths = gate.get_default_test_paths(__file__, "gate_test029_volume_time_rotation")


def create_simulation(sim, aa_flag):
    # main options
    ui = sim.user_info
    ui.g4_verbose = False
    ui.visu = False
    ui.number_of_threads = 1
    ui.random_seed = 123456789

    # units
    m = gate.g4_units("m")
    cm = gate.g4_units("cm")
    cm3 = gate.g4_units("cm3")
    keV = gate.g4_units("keV")
    mm = gate.g4_units("mm")
    Bq = gate.g4_units("Bq")
    sec = gate.g4_units("second")
    BqmL = Bq / cm3

    # world size
    world = sim.world
    world.size = [2 * m, 2 * m, 2 * m]
    world.material = "G4_AIR"

    # spect head (no collimator)
    spect, crystal = gate_spect.add_ge_nm67_spect_head(
        sim, "spect", collimator_type=False, debug=False
    )
    initial_rot = Rotation.from_euler("X", 90, degrees=True)
    t, rot = gate.get_transform_orbiting([0, 25 * cm, 0], "Z", 0)
    rot = Rotation.from_matrix(rot)
    spect.translation = t
    spect.rotation = (rot * initial_rot).as_matrix()

    # iec phantom
    gate_iec.add_phantom(sim)

    # two sources (no background yet)
    activity_concentration = 5000 * BqmL / ui.number_of_threads
    ac = activity_concentration
    sources = gate_iec.add_spheres_sources(
        sim,
        "iec",
        "iec_source",
        [10, 13, 17, 22, 28, 37],
        [ac, ac, ac, ac, ac, ac],
        verbose=True,
    )

    for s in sources:
        s.particle = "gamma"
        s.energy.type = "mono"
        s.energy.mono = 140 * keV
        s.direction.type = "iso"
        s.direction.type = "momentum"
        s.direction.momentum = [0, 1, 0]
        s.direction.acceptance_angle.volumes = ["spect"]
        s.direction.acceptance_angle.intersection_flag = aa_flag
        s.direction.acceptance_angle.skip_policy = "ZeroEnergy"

    sources = gate_iec.add_spheres_sources(
        sim,
        "iec",
        "iec_source2",
        [10, 13, 17, 22, 28, 37],
        [ac, ac, ac, ac, ac, ac],
        verbose=True,
    )

    for s in sources:
        s.particle = "gamma"
        s.energy.type = "mono"
        s.energy.mono = 140 * keV
        s.direction.type = "iso"
        s.direction.type = "momentum"
        s.direction.momentum = [1, 0, 0]
        s.direction.acceptance_angle.volumes = ["spect"]
        s.direction.acceptance_angle.intersection_flag = aa_flag
        s.direction.acceptance_angle.skip_policy = "ZeroEnergy"

    # physic list
    sim.set_physics_list("G4EmStandardPhysics_option4")
    sim.set_cut("world", "all", 10 * mm)
    sim.set_cut("spect", "all", 1 * mm)

    # add stat actor
    stat = sim.add_actor("SimulationStatisticsActor", "Stats")
    stat.output = paths.output / "stats029.txt"

    # hits collection
    hc = sim.add_actor("DigitizerHitsCollectionActor", "Hits")
    hc.mother = "spect_crystal"
    hc.output = ""  # No output
    hc.attributes = [
        "PostPosition",
        "TotalEnergyDeposit",
        "PreStepUniqueVolumeID",
        "GlobalTime",
    ]

    # singles collection
    sc = sim.add_actor("DigitizerAdderActor", "Singles")
    sc.mother = hc.mother
    sc.input_digi_collection = "Hits"
    sc.policy = "EnergyWeightedCentroidPosition"
    sc.output = hc.output

    # EnergyWindows
    cc = sim.add_actor("DigitizerEnergyWindowsActor", "EnergyWindows")
    cc.mother = hc.mother
    cc.input_digi_collection = "Singles"
    cc.channels = [
        {"name": "scatter", "min": 114 * keV, "max": 126 * keV},
        {"name": "peak140", "min": 126 * keV, "max": 154.55 * keV},
    ]
    cc.output = hc.output

    # projections
    proj = sim.add_actor("HitsProjectionActor", "Projection")
    proj.mother = hc.mother
    proj.input_digi_collections = ["Singles", "scatter", "peak140"]
    proj.spacing = [4.41806 * mm, 4.41806 * mm]
    proj.size = [128, 128]
    proj.origin_as_image_center = False
    proj.output = paths.output / "proj029.mhd"

    # motion of the spect, create also the run time interval
    motion = sim.add_actor("MotionVolumeActor", "Move")
    motion.mother = spect.name
    motion.translations = []
    motion.rotations = []
    n = 9
    sim.run_timing_intervals = []
    start = -90
    gantry_rotation = start
    end = 1 * sec / n
    initial_rot = Rotation.from_euler("X", 90, degrees=True)
    for r in range(n):
        t, rot = gate.get_transform_orbiting([0, 30 * cm, 0], "Z", gantry_rotation)
        rot = Rotation.from_matrix(rot)
        rot = rot * initial_rot
        rot = gate.rot_np_as_g4(rot.as_matrix())
        motion.translations.append(t)
        motion.rotations.append(rot)
        sim.run_timing_intervals.append([start, end])
        gantry_rotation += 10
        start = end
        end += 1 * sec / n

    print(f"Run {len(sim.run_timing_intervals)} intervals: {sim.run_timing_intervals}")

    # check actor priority: the MotionVolumeActor must be first
    l = [l for l in sim.actor_manager.user_info_actors.values()]
    sorted_actors = sorted(l, key=lambda d: d.priority)
    print(f"Actors order: ", [[l.name, l.priority] for l in sorted_actors])



from .GenericSource import *
import opengate_core as g4
import itk


class VoxelsSource(GenericSource):
    """
    Voxels source for 3D distributed activity.
    Sampled with cumulative distribution functions.
    """

    type_name = "VoxelsSource"

    @staticmethod
    def set_default_user_info(user_info):
        GenericSource.set_default_user_info(user_info)
        # additional option: image and coord_syst
        user_info.image = None
        # add position translation
        user_info.position = Box()
        user_info.position.translation = [0, 0, 0]
        user_info.position.confine = None
        # no rotation for the moment
        user_info.position.rotation = Rotation.identity().as_matrix()
        # default values
        user_info.direction.type = "iso"
        user_info.energy.type = "mono"
        user_info.energy.mono = 0

    def __del__(self):
        pass

    def __init__(self, user_info):
        super().__init__(user_info)
        self.image = None

    def __getstate__(self):
        super().__getstate__()
        self.image = None
        return self.__dict__

    def create_g4_source(self):
        return g4.GateVoxelsSource()

    def set_transform_from_user_info(self):
        # get source image information
        src_info = gate.read_image_info(str(self.user_info.image))
        # get pointer to SPSVoxelPosDistribution
        pg = self.g4_source.GetSPSVoxelPosDistribution()
        # update cpp image info (no need to allocate)
        gate.update_image_py_to_cpp(self.image, pg.cpp_edep_image, False)
        # set spacing
        pg.cpp_edep_image.set_spacing(src_info.spacing)
        # set origin (half size + translation and half pixel shift)
        c = (
            -src_info.size / 2.0 * src_info.spacing
            + self.user_info.position.translation
            + src_info.spacing / 2.0
        )
        pg.cpp_edep_image.set_origin(c)

    def cumulative_distribution_functions(self):
        """
        Compute the Cumulative Distribution Function of the image
        Composed of: CDF_Z = 1D, CDF_Y = 2D, CDF_X = 3D
        """
        cdf_x, cdf_y, cdf_z = gate.compute_image_3D_CDF(self.image)

        # set CDF to the position generator
        pg = self.g4_source.GetSPSVoxelPosDistribution()
        pg.SetCumulativeDistributionFunction(cdf_z, cdf_y, cdf_x)

    def initialize(self, run_timing_intervals):
        # read source image
        self.image = itk.imread(gate.check_filename_type(self.user_info.image))

        # compute position
        self.set_transform_from_user_info()

        # create Cumulative Distribution Function
        self.cumulative_distribution_functions()

        # initialize standard options (particle energy, etc)
        # we temporarily set the position attribute to reuse
        # the GenericSource verification
        GenericSource.initialize(self, run_timing_intervals)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import itk
from scipy.spatial.transform import Rotation
import gatetools as gt
import numpy as np

paths = gate.get_default_test_paths(__file__, "")

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.visu = False
ui.number_of_threads = 1
ui.random_seed = 123456
print(ui)

# add a material database
sim.add_material_database(paths.data / "GateMaterials.db")

# units
m = gate.g4_units("m")
mm = gate.g4_units("mm")
cm = gate.g4_units("cm")
keV = gate.g4_units("keV")
MeV = gate.g4_units("MeV")
Bq = gate.g4_units("Bq")
kBq = 1000 * Bq

#  change world size
world = sim.world
world.size = [1.5 * m, 1 * m, 1 * m]

# fake box #1
fake = sim.add_volume("Box", "fake")
fake.size = [66 * cm, 66 * cm, 66 * cm]
fake.translation = [25 * cm, 5 * cm, 3 * cm]
r = Rotation.from_euler("z", 5, degrees=True)
r = r * Rotation.from_euler("y", -15, degrees=True)
fake.rotation = r.as_matrix()

# ---------------------------------------------------
# CT image #1
ct = sim.add_volume("Image", "ct")
ct.image = str(paths.data / "empty_anisotrop.mhd")
ct.mother = fake.name
ct.voxel_materials = [[0, 10000000, "G4_WATER"]]  # only water
ct.translation = [-3 * cm, 5 * cm, -2 * cm]
r = Rotation.from_euler("x", 45, degrees=True)
ct.rotation = r.as_matrix()

ct_info = gate.read_image_info(ct.image)
print(f"CT image origin and size: ", ct_info.origin, ct_info.size, ct_info.spacing)

# source from image for CT #1
source = sim.add_source("VoxelsSource", "vox_source")
source.mother = ct.name
source.particle = "alpha"
source.activity = 20000 * Bq / ui.number_of_threads
source.image = str(paths.data / "five_pixels_anisotrop.mhd")
source.direction.type = "iso"
source.position.translation = gate.get_translation_between_images_center(
    ct.image, source.image
)
print(f"Source wrt CT translation", source.position.translation)
source.energy.mono = 1 * MeV
src_info = gate.read_image_info(source.image)
print(
    f"Source image origin and size: ", src_info.origin, src_info.size, src_info.spacing
)

# add dose actor
dose = sim.add_actor("DoseActor", "dose")
dose.output = str(paths.output / "test021-edep_2.mhd")
dose.mother = ct.name
img_info = gate.read_image_info(ct.image)

"""
# test same size/spacing than source : OK
dose.size = src_info.size
dose.spacing = src_info.spacing
"""

"""
# test same different spacing : OK
dose.size = src_info.size * 2
dose.spacing = src_info.spacing / 1.5
"""

# test random size/spacing
dose.size = [25, 18, 30]
dose.spacing = [8.0, 5.0, 10.0]

print(f"dose image:", dose.size)
print(f"dose image:", dose.spacing)
dose.img_coord_system = True

# cuts
p = sim.get_physics_user_info()
p.physics_list_name = "QGSP_BERT_EMZ"
p.enable_decay = False
sim.set_cut("world", "all", 1 * mm)

# add stat actor
stats = sim.add_actor("SimulationStatisticsActor", "Stats")
stats.track_types_flag = True

# verbose
sim.apply_g4_command("/tracking/verbose 0")

# start simulation
output = sim.start()

# print results at the end
stat = output.get_actor("Stats")
# stat.write(paths.output_ref / "stat021_ref_2.txt")

# test pixels in dose #1
final_dose = itk.imread(dose.output)
s = itk.array_view_from_image(final_dose).sum()

# loo for all source pixels (should be five)
src = itk.imread(source.image)
p = []
arr = itk.array_view_from_image(src)
for [i, j, k], flow in np.ndenumerate(src):
    if arr[i, j, k] > 0:
        p.append([k, j, i])
# convert into dose coord system
p_mm = [src.TransformIndexToPhysicalPoint(index) for index in p]
p_d = [final_dose.TransformPhysicalPointToIndex(p) for p in p_mm]
# get the values
v = [final_dose.GetPixel(index) for index in p_d]
print(v)

is_ok = True
tol = 208
for vv, pp in zip(v, p):
    b = vv > tol
    gate.print_test(b, f"Compare value at {pp} : {vv:.2f} > {tol}] ?  {b}")
    is_ok = is_ok and b

stats_ref = gate.read_stat_file(paths.output_ref / "stat021_ref_2.txt")
stats_ref.counts.run_count = ui.number_of_threads
is_ok = gate.assert_stats(stat, stats_ref, 0.1) and is_ok

is_ok = (
    gate.assert_images(
        paths.output_ref / "test021-edep_2.mhd",
        dose.output,
        stat,
        tolerance=11,
        ignore_value=0,
        axis="y",
    )
    and is_ok
)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate_core as g4

paths = gate.get_default_test_paths(__file__, "gate_test025_hits_collection")


def create_simulation(nb_threads):
    # create the simulation
    sim = gate.Simulation()

    # main options
    ui = sim.user_info
    ui.g4_verbose = False
    ui.visu = False
    ui.number_of_threads = nb_threads
    ui.check_volumes_overlap = False

    # units
    m = gate.g4_units("m")
    cm = gate.g4_units("cm")
    keV = gate.g4_units("keV")
    mm = gate.g4_units("mm")
    Bq = gate.g4_units("Bq")

    # world size
    world = sim.world
    world.size = [2 * m, 2 * m, 2 * m]

    # material
    sim.add_material_database(paths.data / "GateMaterials.db")

    # fake spect head
    waterbox = sim.add_volume("Box", "SPECThead")
    waterbox.size = [55 * cm, 42 * cm, 18 * cm]
    waterbox.material = "G4_AIR"

    # crystal
    crystal1 = sim.add_volume("Box", "crystal1")
    crystal1.mother = "SPECThead"
    crystal1.size = [0.5 * cm, 0.5 * cm, 2 * cm]
    crystal1.translation = None
    crystal1.rotation = None
    crystal1.material = "NaITl"
    start = [-25 * cm, -20 * cm, 4 * cm]
    size = [100, 40, 1]
    # size = [100, 80, 1]
    tr = [0.5 * cm, 0.5 * cm, 0]
    crystal1.repeat = gate.repeat_array_start("crystal1", start, size, tr)
    crystal1.color = [1, 1, 0, 1]

    # additional volume
    crystal2 = sim.add_volume("Box", "crystal2")
    crystal2.mother = "SPECThead"
    crystal2.size = [0.5 * cm, 0.5 * cm, 2 * cm]
    crystal2.translation = None
    crystal2.rotation = None
    crystal2.material = "NaITl"
    start = [-25 * cm, 0 * cm, 4 * cm]
    size = [100, 40, 1]
    tr = [0.5 * cm, 0.5 * cm, 0]
    crystal2.repeat = gate.repeat_array_start("crystal2", start, size, tr)
    crystal2.color = [0, 1, 1, 1]

    # physic list
    p = sim.get_physics_user_info()
    p.physics_list_name = "G4EmStandardPhysics_option4"
    p.enable_decay = False
    cuts = p.production_cuts
    cuts.world.gamma = 0.01 * mm
    cuts.world.electron = 0.01 * mm
    cuts.world.positron = 1 * mm
    cuts.world.proton = 1 * mm

    # default source for tests
    source = sim.add_source("GenericSource", "Default")
    source.particle = "gamma"
    source.energy.mono = 140.5 * keV
    source.position.type = "sphere"
    source.position.radius = 4 * cm
    source.position.translation = [0, 0, -15 * cm]
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.activity = 50000 * Bq / ui.number_of_threads

    # add stat actor
    sim.add_actor("SimulationStatisticsActor", "Stats")

    # print list of attributes
    am = g4.GateDigiAttributeManager.GetInstance()
    print(am.GetAvailableDigiAttributeNames())

    # hits collection
    hc = sim.add_actor("DigitizerHitsCollectionActor", "Hits")
    hc.mother = [crystal1.name, crystal2.name]
    mt = ""
    if ui.number_of_threads > 1:
        mt = "_MT"
    hc.output = paths.output / ("test025_hits" + mt + ".root")
    hc.attributes = [
        "TotalEnergyDeposit",
        "KineticEnergy",
        "PostPosition",
        "TrackCreatorProcess",
        "GlobalTime",
        "TrackVolumeName",
        "RunID",
        "ThreadID",
        "TrackID",
    ]

    """
    ## NO DYNAMIC BRANCH YET --> bug in MT mode
    # dynamic branch creation (SLOW !)
    def branch_fill(att, step, touchable):
        e = step.GetTotalEnergyDeposit()
        att.FillDValue(e)
        # branch.push_back_double(e)
        # branch.push_back_double(123.3)
        # print('done')

    # dynamic branch
    man = gate_g4.GateDigiAttributeManager.GetInstance()
    man.DefineDigiAttribute('MyBranch', 'D', branch_fill)
    #hc.attributes.append('MyBranch')
    """

    print("List of active attributes (including dynamic attributes)", hc.attributes)

    # hits collection #2
    hc2 = sim.add_actor("DigitizerHitsCollectionActor", "Hits2")
    hc2.mother = [crystal1.name, crystal2.name]
    hc2.output = paths.output / ("test025_secondhits" + mt + ".root")
    hc2.attributes = ["TotalEnergyDeposit", "GlobalTime"]

    # --------------------------------------------------------------------------------------------------
    # create G4 objects
    sec = gate.g4_units("second")
    sim.run_timing_intervals = [
        [0, 0.15 * sec],
        [0.15 * sec, 0.16 * sec],
        [0.16 * sec, 1 * sec],
    ]
    # sim.run_timing_intervals = [[0, 1 * sec]]
    # sim.run_timing_intervals = [[0, 0.5 * sec], [0.5 * sec, 1 * sec]]

    # ui.running_verbose_level = gate.EVENT
    return sim


def test_simulation_results(sim, output):
    # Compare stats file
    stats = output.get_actor("Stats")
    print(f"Number of runs was {stats.counts.run_count}. Set to 1 before comparison")
    stats.counts.run_count = 1  # force to 1 to compare with gate result
    stats_ref = gate.read_stat_file(paths.gate_output / "stat.txt")
    is_ok = gate.assert_stats(stats, stats_ref, tolerance=0.05)

    # Compare root files
    print()
    gate_file = paths.gate_output / "hits.root"
    hc_file = output.get_actor("Hits").user_info.output
    checked_keys = ["posX", "posY", "posZ", "edep", "time", "trackId"]
    keys1, keys2, scalings, tols = gate.get_keys_correspondence(checked_keys)
    # tols[0] = 0.97   # PostPosition_X
    tols[3] = 0.002  # edep
    is_ok = (
        gate.compare_root3(
            gate_file,
            hc_file,
            "Hits",
            "Hits",
            keys1,
            keys2,
            tols,
            [1] * len(scalings),
            scalings,
            paths.output / "test025.png",
        )
        and is_ok
    )

    """# compare the dynamic branch
    print()
    t1 = uproot.open(gate_file)['Hits'].arrays(library="numpy")
    t2 = uproot.open(hc_file)['Hits'].arrays(library="numpy")
    gate.compare_branches(t1, list(t1.keys()), t2, list(t2.keys()),
                         'edep', 'MyBranch', 0.2, 1.0, False)"""

    # Compare root files
    print()
    gate_file = paths.gate_output / "hits.root"
    hc_file = output.get_actor("Hits2").user_info.output
    checked_keys = ["time", "edep"]
    keys1, keys2, scalings, tols = gate.get_keys_correspondence(checked_keys)
    tols[1] = 0.002  # edep
    is_ok = (
        gate.compare_root3(
            gate_file,
            hc_file,
            "Hits",
            "Hits2",
            keys1,
            keys2,
            tols,
            [1] * len(scalings),
            scalings,
            paths.output / "test025_secondhits.png",
        )
        and is_ok
    )

    # this is the end, my friend
    gate.test_ok(is_ok)



import opengate as gate
from scipy.spatial.transform import Rotation
from box import Box


def create_material(name):
    gcm3 = gate.g4_units("g/cm3")
    gate.new_material_nb_atoms(
        f"{name}_target_tungsten", 19.4 * gcm3, ["W", "Re"], [9, 1]
    )
    gate.new_material_weights(f"{name}_target_copper", 8.93 * gcm3, "Cu")
    gate.new_material_weights(
        f"{name}_colli", 18 * gcm3, ["W", "Ni", "Fe"], [0.95, 0.0375, 0.0125]
    )
    gate.new_material_weights(
        f"{name}_flat_filter", 7.9 * gcm3, ["Cr", "Fe", "Ni"], [0.17, 0.75, 0.08]
    )
    gate.new_material_weights(
        f"{name}_mylar", 1.38 * gcm3, ["H", "C", "O"], [0.04196, 0.625016, 0.333024]
    )
    gate.new_material_weights(f"{name}_carbon", 2.27 * gcm3, "C")
    gate.new_material_weights(f"{name}_aluminium", 2.7 * gcm3, "Al")


def add_linac(sim, name="linac"):
    # unit
    m = gate.g4_units("m")
    mm = gate.g4_units("mm")
    deg = gate.g4_units("deg")
    create_material(name)

    # for debug : should be the same as create_material
    # sim.add_material_database('../contrib/linac_elekta_synergy_materials.db')

    # colors
    red = [1, 0.7, 0.7, 0.8]
    blue = [0.5, 0.5, 1, 0.8]
    white = [1, 1, 1, 0.8]

    # check overlap
    sim.g4_check_overlap_flag = True

    # global box
    linac = sim.add_volume("Box", name)
    linac.material = "G4_AIR"
    linac.size = [0.25 * m, 0.25 * m, 0.6 * m]
    linac.color = white

    # target
    add_target(sim, linac.name)

    # primary collimator
    primary_collimator = sim.add_volume("Cons", f"{name}_primary_collimator")
    primary_collimator.mother = linac.name
    primary_collimator.material = f"{name}_colli"
    primary_collimator.rmin1 = 31.45 * mm
    primary_collimator.rmax1 = 82 * mm
    primary_collimator.rmin2 = 6.45 * mm
    primary_collimator.rmax2 = 82 * mm
    primary_collimator.dz = 101 * mm / 2.0
    primary_collimator.sphi = 0
    primary_collimator.dphi = 360 * deg
    primary_collimator.translation = [0, 0, -65.5 * mm]
    primary_collimator.color = blue

    # flattening_filter
    add_flattening_filter(sim, linac.name)

    # ionizing_chamber
    add_ionizing_chamber(sim, linac.name)

    # back_scatter_plate
    bsp = sim.add_volume("Box", f"{name}_back_scatter_plate")
    bsp.mother = linac.name
    bsp.material = f"{name}_aluminium"
    bsp.size = [116 * mm, 84 * mm, 3 * mm]
    bsp.translation = [0, 0, -183 * mm]
    bsp.color = red

    # mirror
    add_mirror(sim, linac.name)

    return linac


def add_target(sim, name):
    # unit
    mm = gate.g4_units("mm")

    # colors
    red = [1, 0.2, 0.2, 0.8]
    green = [0, 1, 0, 0.2]

    # material
    target_material = f"{name}_target_tungsten"
    copper = f"{name}_target_copper"

    # target
    target_support = sim.add_volume("Tubs", f"{name}_target_support")
    target_support.mother = name
    target_support.material = "G4_AIR"
    target_support.rmin = 0
    target_support.rmax = 15 * mm
    target_support.dz = 11 * mm / 2.0
    target_support.translation = [0, 0, -5 * mm]
    target_support.color = [0, 0, 0, 0]  # invisible

    target = sim.add_volume("Tubs", f"{name}_target")
    target.mother = target_support.name
    target.material = target_material
    target.rmin = 0
    target.rmax = 2.7 * mm
    target.dz = 1 * mm / 2.0
    target.translation = [0, 0, 5 * mm]
    target.color = red

    target_support_top = sim.add_volume("Tubs", f"{name}_target_support_top")
    target_support_top.mother = target_support.name
    target_support_top.material = copper
    target_support_top.rmin = 2.7 * mm
    target_support_top.rmax = 15 * mm
    target_support_top.dz = 1 * mm / 2.0
    target_support_top.translation = [0, 0, 5 * mm]
    target_support_top.color = green

    target_support_bottom = sim.add_volume("Tubs", f"{name}_target_support_bottom")
    target_support_bottom.mother = target_support.name
    target_support_bottom.material = copper
    target_support_bottom.rmin = 0
    target_support_bottom.rmax = 15 * mm
    target_support_bottom.dz = 10 * mm / 2.0
    target_support_bottom.translation = [0, 0, -0.5 * mm]
    target_support_bottom.color = green


def add_flattening_filter(sim, name):
    # unit
    mm = gate.g4_units("mm")
    deg = gate.g4_units("deg")

    # colors
    red = [1, 0.7, 0.7, 0.8]
    green = [0.5, 1, 0.5, 0.8]
    yellow = [0, 0.7, 0.7, 0.8]

    # bounding cylinder
    flattening_filter = sim.add_volume("Tubs", f"{name}_flattening_filter")
    flattening_filter.mother = name
    flattening_filter.material = "G4_AIR"
    flattening_filter.rmin = 0
    flattening_filter.rmax = 40 * mm
    flattening_filter.dz = 24.1 * mm / 2
    flattening_filter.translation = [0, 0, -146.05 * mm]
    flattening_filter.color = [0, 0, 0, 0]  # invisible

    # create all cones
    def add_cone(sim, p):
        c = sim.add_volume("Cons", f"{name}_flattening_filter_cone_{p.name}")
        c.mother = f"{name}_flattening_filter"
        c.material = f"{name}_flat_filter"
        c.rmin1 = 0
        c.rmax1 = p.rmax1
        c.rmin2 = 0
        c.rmax2 = p.rmax2
        c.dz = p.dz
        c.sphi = 0
        c.dphi = 360 * deg
        c.translation = [0, 0, p.tr]
        c.color = yellow

    cones = [
        [0.001, 5.45, 3.40, 10.35],
        [5.45, 9, 2.7, 7.3],
        [9, 14.5, 4.9, 3.5],
        [14.5, 22.5, 5.5, -1.7],
        [22.5, 32.5, 5.6, -7.25],
        [38.5, 38.5, 2, -11.05],
    ]  ## FIXME check 32.5 ?
    i = 0
    for c in cones:
        cone = Box()
        cone.name = i
        cone.rmax2 = c[0] * mm
        cone.rmax1 = c[1] * mm
        cone.dz = c[2] * mm / 2  # /2 to keep same values than Gate (where dz was /2)
        cone.tr = c[3] * mm
        add_cone(sim, cone)
        i = i + 1


def add_ionizing_chamber(sim, name):
    # unit
    mm = gate.g4_units("mm")

    # main cylinder
    ionizing_chamber = sim.add_volume("Tubs", f"{name}_ionizing_chamber")
    ionizing_chamber.mother = name
    ionizing_chamber.material = "G4_AIR"
    ionizing_chamber.rmin = 0
    ionizing_chamber.rmax = 45 * mm
    ionizing_chamber.dz = 9.28 * mm / 2
    ionizing_chamber.translation = [0, 0, -169 * mm]
    ionizing_chamber.color = [0, 0, 0, 0]

    # layers
    def add_layer(sim, p):
        l = sim.add_volume("Tubs", f"{name}_ionizing_chamber_mylar_layer_{p.i}")
        l.mother = f"{name}_ionizing_chamber"
        l.material = f"{name}_mylar"
        l.rmin = 0
        l.rmax = 45 * mm
        l.dz = 0.012 * mm / 2
        l.translation = [0, 0, p.tr1]

        l = sim.add_volume("Tubs", f"{name}_ionizing_chamber_carbon_layer_{p.i}")
        l.mother = f"{name}_ionizing_chamber"
        l.material = f"{name}_carbon"
        l.rmin = 0
        l.rmax = 45 * mm
        l.dz = 0.000150 * mm / 2
        l.translation = [0, 0, p.tr2]

    layers = [
        [-2.634, -2.627925],
        [-0.434, -0.427925],
        [0.566, 0.572075],
        [1.566, 1.572075],
        [2.566, 2.572075],
        [3.566, 3.572075],
    ]
    i = 1
    for l in layers:
        ll = Box()
        ll.i = i
        ll.tr1 = l[0] * mm
        ll.tr2 = l[1] * mm
        add_layer(sim, ll)
        i = i + 1


def add_mirror(sim, name):
    # unit
    mm = gate.g4_units("mm")
    blue = [0, 0, 1, 0.8]

    # main box
    m = sim.add_volume("Box", f"{name}_mirror")
    m.mother = name
    m.material = "G4_AIR"
    m.size = [137 * mm, 137 * mm, 1.5 * mm]
    m.translation = [0, 0, -225 * mm]
    rot = Rotation.from_euler("x", 37.5, degrees=True)
    m.rotation = rot.as_matrix()

    # mylar
    l = sim.add_volume("Box", f"{name}_mirror_mylar_layer")
    l.mother = m.name
    l.material = f"{name}_mylar"
    l.size = [110 * mm, 110 * mm, 0.0012 * mm]
    l.translation = [0, 0, 0.15 * mm]
    l.color = blue

    # alu
    l = sim.add_volume("Box", f"{name}_mirror_alu_layer")
    l.mother = m.name
    l.material = f"{name}_aluminium"
    l.size = [110 * mm, 110 * mm, 0.0003 * mm]
    l.translation = [0, 0, -0.6 * mm]
    l.color = blue



import opengate as gate
from .ARFActor import *
from .ARFTrainingDatasetActor import *
from .DoseActor import *
from .LETActor import *
from .DigitizerAdderActor import *
from .DigitizerReadoutActor import *
from .DigitizerEnergyWindowsActor import *
from .DigitizerProjectionActor import *
from .DigitizerBlurringActor import *
from .DigitizerSpatialBlurringActor import *
from .MotionVolumeActor import *
from .PhaseSpaceActor import *
from .SimulationStatisticsActor import *
from .SourceInfoActor import *
from .TestActor import *

actor_type_names = {
    SimulationStatisticsActor,
    DoseActor,
    LETActor,
    SourceInfoActor,
    PhaseSpaceActor,
    DigitizerHitsCollectionActor,
    DigitizerAdderActor,
    DigitizerEnergyWindowsActor,
    DigitizerProjectionActor,
    DigitizerReadoutActor,
    DigitizerBlurringActor,
    DigitizerSpatialBlurringActor,
    MotionVolumeActor,
    ARFActor,
    ARFTrainingDatasetActor,
    TestActor,
}
actor_builders = gate.make_builders(actor_type_names)


def get_simplified_digitizer_channels_Tc99m(spect_name, scatter_flag):
    keV = gate.g4_units("keV")
    # Tc99m
    channels = [
        {"name": f"scatter_{spect_name}", "min": 114 * keV, "max": 126 * keV},
        {"name": f"peak140_{spect_name}", "min": 126 * keV, "max": 154 * keV},
    ]
    if not scatter_flag:
        channels.pop(0)
    return channels


def get_simplified_digitizer_channels_Lu177(spect_name, scatter_flag):
    # Lu177, Ljungberg2016
    keV = gate.g4_units("keV")
    channels = [
        {"name": f"scatter1_{spect_name}", "min": 96 * keV, "max": 104 * keV},
        {"name": f"peak113_{spect_name}", "min": 104 * keV, "max": 121.48 * keV},
        {"name": f"scatter2_{spect_name}", "min": 122.48 * keV, "max": 133.12 * keV},
        {"name": f"scatter3_{spect_name}", "min": 176.46 * keV, "max": 191.36 * keV},
        {"name": f"peak208_{spect_name}", "min": 192.36 * keV, "max": 223.6 * keV},
    ]
    if not scatter_flag:
        channels.pop(0)
        channels.pop(1)
        channels.pop(1)
    return channels


def get_simplified_digitizer_channels_In111(spect_name, scatter_flag):
    # In111
    keV = gate.g4_units("keV")
    channels = [
        {"name": f"scatter1_{spect_name}", "min": 150 * keV, "max": 156 * keV},
        {"name": f"peak171_{spect_name}", "min": 156 * keV, "max": 186 * keV},
        {"name": f"scatter2_{spect_name}", "min": 186 * keV, "max": 192 * keV},
        {"name": f"scatter3_{spect_name}", "min": 218 * keV, "max": 224 * keV},
        {"name": f"peak245_{spect_name}", "min": 224 * keV, "max": 272 * keV},
    ]
    if not scatter_flag:
        channels.pop(0)
        channels.pop(1)
        channels.pop(1)
    return channels


def get_simplified_digitizer_channels_I131(spect_name, scatter_flag):
    # I131
    keV = gate.g4_units("keV")
    channels = [
        {"name": f"scatter1_{spect_name}", "min": 314 * keV, "max": 336 * keV},
        {"name": f"peak364_{spect_name}", "min": 336 * keV, "max": 392 * keV},
        {"name": f"scatter2_{spect_name}", "min": 392 * keV, "max": 414 * keV},
        {"name": f"scatter3_{spect_name}", "min": 414 * keV, "max": 556 * keV},
        {"name": f"scatter4_{spect_name}", "min": 556 * keV, "max": 595 * keV},
        {"name": f"peak637_{spect_name}", "min": 595 * keV, "max": 679 * keV},
    ]
    if not scatter_flag:
        channels.pop(0)
        channels.pop(1)
        channels.pop(1)
        channels.pop(1)
    return channels


def get_simplified_digitizer_channels_rad(spect_name, rad, scatter_flag):
    available_rad = {
        "Tc99m": get_simplified_digitizer_channels_Tc99m,
        "Lu177": get_simplified_digitizer_channels_Lu177,
        "In111": get_simplified_digitizer_channels_In111,
        "I131": get_simplified_digitizer_channels_I131,
    }

    if rad not in available_rad:
        gate.fatal(
            f"Error, the radionuclide {rad} is not known, list of available is: {available_rad}"
        )

    return available_rad[rad](spect_name, scatter_flag)
